# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A brax environment for training and inference."""

import abc
from typing import Any, Dict, Optional

import brax
from brax import jumpy as jp
from brax import pytree
from flax import struct

from google.protobuf import text_format


@struct.dataclass
class State:
  """Environment state for training and inference."""
  qp: brax.QP
  obs: jp.ndarray
  reward: jp.ndarray
  done: jp.ndarray
  metrics: Dict[str, jp.ndarray] = struct.field(default_factory=dict)
  info: Dict[str, Any] = struct.field(default_factory=dict)


@pytree.register
class Env(abc.ABC):
  """API for driving a brax system for training and inference."""


  def __init__(self, config: Optional[str], *args, **kwargs):
    if config:
      config = text_format.Parse(config, brax.Config())
      self.sys = brax.System(config, *args, **kwargs)

  @abc.abstractmethod
  def reset(self, rng: jp.ndarray) -> State:
    """Resets the environment to an initial state."""

  @abc.abstractmethod
  def step(self, state: State, action: jp.ndarray) -> State:
    """Run one timestep of the environment's dynamics."""

  @property
  def observation_size(self) -> int:
    """The size of the observation vector returned in step and reset."""
    rng = jp.random_prngkey(0)
    reset_state = self.unwrapped.reset(rng)
    return reset_state.obs.shape[-1]

  @property
  def action_size(self) -> int:
    """The size of the action vector expected by step."""
    return self.sys.num_joint_dof + self.sys.num_forces_dof

  @property
  def unwrapped(self) -> 'Env':
    return self


class Wrapper(Env):
  """Wraps the environment to allow modular transformations."""

  def __init__(self, env: Env):
    super().__init__(config=None)
    self.env = env

  def reset(self, rng: jp.ndarray) -> State:
    return self.env.reset(rng)

  def step(self, state: State, action: jp.ndarray) -> State:
    return self.env.step(state, action)

  @property
  def observation_size(self) -> int:
    return self.env.observation_size

  @property
  def action_size(self) -> int:
    return self.env.action_size

  @property
  def unwrapped(self) -> Env:
    return self.env.unwrapped

  def __getattr__(self, name):
    if name == '__setstate__':
      raise AttributeError(name)
    return getattr(self.env, name)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Environments for training and evaluating policies."""

from typing import Optional, Type

from brax.v2.envs import ant
from brax.v2.envs import wrapper
from brax.v2.envs.env import Env


_envs = {
    'ant': ant.Ant,
}


def get_environment(env_name: str, **kwargs) -> Env:
  """Returns an environment from the environment registry.

  Args:
    env_name: environment name string
    **kwargs: keyword arguments that get passed to the Env class constructor

  Returns:
    env: an environment
  """
  return _envs[env_name](**kwargs)


def register_environment(env_name: str, env_class: Type[Env]):
  """Adds an environment to the registry.

  Args:
    env_name: environment name string
    env_class: the Env class to add to the registry
  """
  _envs[env_name] = env_class


def create(
    env_name: str,
    episode_length: int = 1000,
    action_repeat: int = 1,
    auto_reset: bool = True,
    batch_size: Optional[int] = None,
    **kwargs,
) -> Env:
  """Creates an environment from the registry.

  Args:
    env_name: environment name string
    episode_length: length of episode
    action_repeat: how many repeated actions to take per environment step
    auto_reset: whether to auto reset the environment after an episode is done
    batch_size: the number of environments to batch together
    **kwargs: keyword arugments that get passed to the Env class constructor

  Returns:
    env: an environment
  """
  env = _envs[env_name](**kwargs)

  if episode_length is not None:
    env = wrapper.EpisodeWrapper(env, episode_length, action_repeat)
  if batch_size:
    env = wrapper.VmapWrapper(env, batch_size)
  if auto_reset:
    env = wrapper.AutoResetWrapper(env)

  return env  # type: ignore



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a reacher to reach a target.

Based on the OpenAI Gym MuJoCo Reacher environment.
"""

from typing import Tuple

import brax
from brax import jumpy as jp
from brax.envs import env


class ReacherAngle(env.Env):
  """Trains a reacher arm to touch a random target via angle actuators."""

  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)
    self.target_idx = self.sys.body.index['target']
    self.arm_idx = self.sys.body.index['body1']

    # map the [-1, 1] action space into a valid angle for the actuators
    limits = []
    for j in self.sys.config.joints:
      for l in j.angle_limit:
        limits.append((l.min, l.max))
    self._min_act = jp.array([l[0] for l in limits])
    self._range_act = jp.array([l[1] - l[0] for l in limits])

  def reset(self, rng: jp.ndarray) -> env.State:
    rng, rng1, rng2 = jp.random_split(rng, 3)
    qpos = self.sys.default_angle() + jp.random_uniform(
        rng1, (self.sys.num_joint_dof,), -.1, .1)
    qvel = jp.random_uniform(rng2, (self.sys.num_joint_dof,), -.005, .005)
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    rng, target = self._random_target(rng)
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'rewardDist': zero,
        'rewardCtrl': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    action = self._min_act + self._range_act * ((action + 1) / 2.)

    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    # vector from tip to target is last 3 entries of obs vector
    reward_dist = -jp.norm(obs[-3:])

    reward = reward_dist

    metrics = {
        'rewardDist': reward_dist,
        'rewardCtrl': 0.,
    }

    return state.replace(qp=qp, obs=obs, reward=reward, metrics=metrics)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Egocentric observation of target and arm body."""

    # some pre-processing to pull joint angles and velocities
    joint_angle, _ = self.sys.joints[0].angle_vel(qp)

    # qpos:
    # x,y coord of target
    qpos = [qp.pos[self.target_idx, :2]]

    # dist to target and speed of tip
    arm_qps = jp.take(qp, jp.array(self.arm_idx))
    tip_pos, tip_vel = arm_qps.to_world(jp.array([0.11, 0., 0.]))
    tip_to_target = [tip_pos - qp.pos[self.target_idx]]
    cos_sin_angle = [jp.cos(joint_angle), jp.sin(joint_angle)]

    # qvel:
    # velocity of tip
    qvel = [tip_vel[:2]]

    return jp.concatenate(cos_sin_angle + qpos + qvel + tip_to_target)

  def _random_target(self, rng: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns a target location in a random circle slightly above xy plane."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    # we perform a sqrt here so that the final distribution of sampled targets
    # is uniformly random in 2D
    dist = .2 * jp.sqrt(jp.random_uniform(rng1))
    ang = jp.pi * 2. * jp.random_uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    target_z = .01
    target = jp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
  bodies {
    name: "ground"
    colliders {
      plane {
      }
    }
    mass: 1.0
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      all: true
    }
  }
  bodies {
    name: "body0"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "body1"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    colliders {
      position { x: .11 }
      sphere {
        radius: 0.01
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "target"
    colliders {
      position {
      }
      sphere {
        radius: 0.009
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  joints {
    name: "joint0"
    parent: "ground"
    child: "body0"
    parent_offset {
      z: 0.01
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
        min: -180
        max: 180
      }
    angular_damping: 10.0
  }
  joints {
    name: "joint1"
    parent: "body0"
    child: "body1"
    parent_offset {
      x: 0.1
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -180
      max: 180
    }
    angular_damping: 10.0
  }
  actuators {
    name: "joint0"
    joint: "joint0"
    strength: 20.0
    angle {
    }
  }
  actuators {
    name: "joint1"
    joint: "joint1"
    strength: 20.0
    angle {
    }
  }
  collide_include {
  }
  gravity {
    z: -9.81
  }
  dt: 0.02
  substeps: 4
  frozen {
    position {
      z: 1.0
    }
    rotation {
      x: 1.0
      y: 1.0
    }
  }
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "ground"
    colliders {
      plane {
      }
    }
    mass: 1.0
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      all: true
    }
  }
  bodies {
    name: "body0"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "body1"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    colliders {
      position { x: .11 }
      sphere {
        radius: 0.01
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "target"
    colliders {
      position {
      }
      sphere {
        radius: 0.009
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  joints {
    name: "joint0"
    stiffness: 100.0
    parent: "ground"
    child: "body0"
    parent_offset {
      z: 0.01
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
        min: -180
        max: 180
      }
    limit_strength: 0.0
    spring_damping: 3.0
    angular_damping: 10.0
  }
  joints {
    name: "joint1"
    stiffness: 100.0
    parent: "body0"
    child: "body1"
    parent_offset {
      x: 0.1
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -180
      max: 180
    }
    limit_strength: 0.0
    spring_damping: 3.0
    angular_damping: 10.0
  }
  actuators {
    name: "joint0"
    joint: "joint0"
    strength: 20.0
    angle {
    }
  }
  actuators {
    name: "joint1"
    joint: "joint1"
    strength: 20.0
    angle {
    }
  }
  collide_include {
  }
  gravity {
    z: -9.81
  }
  baumgarte_erp: 0.1
  dt: 0.02
  substeps: 4
  frozen {
    position {
      z: 1.0
    }
    rotation {
      x: 1.0
      y: 1.0
    }
  }
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""single-agent environments."""
from brax.experimental.composer import composer_utils

ENV_DESCS = {
    'ant_run':
        dict(
            components=dict(
                agent1=dict(
                    component='ant',
                    pos=(0, 0, 0),
                    reward_fns=dict(
                        goal=dict(
                            reward_type='root_goal',
                            sdcomp='vel',
                            indices=(0, 1),
                            offset=5,
                            target_goal=(4, 0))),
                ),)),
    'ant_chase':
        dict(
            components=dict(
                agent1=dict(component='ant', pos=(0, 0, 0)),
                agent2=dict(
                    component='ant',
                    pos=(0, 2, 0),
                    reward_fns=dict(
                        goal=dict(
                            reward_type='root_goal',
                            sdcomp='vel',
                            indices=(0, 1),
                            offset=5,
                            scale=1,
                            target_goal=(4, 0)),),
                ),
            ),
            edges=dict(
                agent1__agent2=dict(
                    extra_observers=[
                        dict(observer_type='root_vec', indices=(0, 1)),
                    ],
                    reward_fns=dict(
                        dist=dict(
                            reward_type='root_dist', min_dist=1, offset=5)),
                ),)),
    'ant_push':
        dict(
            components=dict(
                agent1=dict(
                    component='ant',
                    pos=(0, 0, 0),
                ),
                cap1=dict(
                    component='singleton',
                    component_params=dict(size=0.5),
                    pos=(1, 0, 0),
                    observers=('root_z_joints',),
                    reward_fns=dict(
                        goal=dict(
                            reward_type='root_goal',
                            sdcomp='vel',
                            indices=(0, 1),
                            offset=5,
                            scale=1,
                            target_goal=5)),
                ),
            ),
            edges=dict(
                agent1__cap1=dict(
                    extra_observers=[
                        dict(observer_type='root_vec', indices=(0, 1)),
                    ],
                    reward_fns=dict(
                        dist=dict(reward_type='root_dist', offset=5)),
                ),)),
    'ant_on_ball':
        dict(
            global_options=dict(dt=0.02, substeps=16),
            components=dict(
                agent1=dict(
                    component='pro_ant',
                    component_params=dict(num_legs=4),
                    pos=(0, 0, 6),
                    term_params=dict(z_offset=6),
                    reward_fns=dict(
                        goal=dict(
                            reward_type='root_goal',
                            sdcomp='vel',
                            indices=(0, 1),
                            offset=4,
                            target_goal=(3, 0))),
                ),
                cap1=dict(
                    component='singleton',
                    component_params=dict(size=3),
                    pos=(0, 0, 0),
                    observers=('root_z_joints',),
                ),
            ),
            edges=dict(
                agent1__cap1=dict(
                    extra_observers=[
                        dict(observer_type='root_vec', indices=(0, 1)),
                    ],),),
        )
}


def create_pro_ant_run(num_legs: int = 4):
  return composer_utils.edit_desc(
      ENV_DESCS['ant_run'], {
          'components.agent1.component': 'pro_ant',
          'components.agent1.component_params': dict(num_legs=num_legs),
          'global_options.dt': 0.02,
          'global_options.substeps': 16,
      })

ENV_DESCS['pro_ant_run'] = create_pro_ant_run

VARIANTS = ()
for comp in ('octopus', 'humanoid', 'halfcheetah'):
  VARIANTS += (('ant_run', f'{comp}_run', {
      'components.agent1.component': comp,
      'global_options.dt': 0.02,
      'global_options.substeps': 16,
  }),)

for base_desc_name, new_desc_name, desc_edits in VARIANTS:
  ENV_DESCS[new_desc_name] = composer_utils.edit_desc(ENV_DESCS[base_desc_name],
                                                      desc_edits)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Colliders push apart bodies that are in contact."""

import abc
import functools
from typing import Any, Callable, List, Optional, Tuple

from brax import jumpy as jp
from brax import math
from brax import pytree
from brax.experimental.tracing import customize
from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import geometry
from brax.physics.base import P, Q, QP, vec_to_arr


class Cull(abc.ABC):
  """Selects collidable pair candidates for collision detection."""

  @abc.abstractmethod
  def get(self) -> Tuple[geometry.Collidable, geometry.Collidable]:
    """Returns collidable pair candidates for collision detection."""

  def update(self, qp: QP):
    """Updates candidate statistics given qp world state."""


@pytree.register
class Pairs(Cull):
  """Naive strategy: returns pairs provided manually."""

  def __init__(self, col_a: geometry.Collidable, col_b: geometry.Collidable):
    self.col_a = col_a
    self.col_b = col_b

  def get(self) -> Tuple[geometry.Collidable, geometry.Collidable]:
    return self.col_a, self.col_b


@pytree.register
class NearNeighbors(Cull):
  """Returns top K nearest neighbor collidables for collision detection."""

  __pytree_ignore__ = ('cutoff',)

  def __init__(self, col_a: geometry.Collidable, col_b: geometry.Collidable,
               mask: Tuple[jp.ndarray, jp.ndarray], cutoff: int):

    dist_off = jp.zeros(col_a.body.idx.shape + col_b.body.idx.shape)
    # TODO: consider accounting for bounds/radius of a collidable
    dist_mask = dist_off + float('inf')
    dist_off += jp.index_update(dist_mask, mask, 0)
    self.dist_off = dist_off
    self.cutoff = cutoff
    self.candidate_a, self.candidate_b = col_a, col_b
    self.col_a, self.col_b = self.candidate_a, self.candidate_b

  def update(self, qp: QP):
    if self.candidate_a is self.candidate_b:
      pos = self.candidate_a.position(qp)
      dist = jp.vmap(lambda pt: jp.norm(pos - pt, axis=-1))(pos)
    else:
      pos_a = self.candidate_a.position(qp)
      pos_b = self.candidate_b.position(qp)
      dist = jp.vmap(lambda pt: jp.norm(pos_b - pt, axis=-1))(pos_a)
    # add in offset and convert to closeness
    sim = -(dist + self.dist_off)
    # TODO: return a status if more valid candidates than cutoff
    _, idx = jp.top_k(sim.ravel(), self.cutoff)
    self.col_a = jp.take(self.candidate_a, idx // sim.shape[-1])
    self.col_b = jp.take(self.candidate_b, idx % sim.shape[-1])

  def get(self) -> Tuple[geometry.Collidable, geometry.Collidable]:
    return self.col_a, self.col_b


class Collider(abc.ABC):
  """Calculates impulses given contacts from a contact function."""

  __pytree_ignore__ = ('contact_fn', 'cull', 'baumgarte_erp', 'collide_scale',
                       'velocity_threshold')

  def __init__(self, contact_fn: Callable[[Any, Any, QP, QP], geometry.Contact],
               cull: Cull, config: config_pb2.Config):
    """Creates a PairwiseCollider that exhaustively checks for contacts.

    Args:
      contact_fn: a function that returns contacts given collidables and qp
      cull: a cull strategy
      config: for specifying global system config
    """
    self.contact_fn = contact_fn
    self.cull = cull
    self.baumgarte_erp = config.baumgarte_erp * config.substeps / config.dt
    self.h = config.dt / config.substeps
    self.substeps = config.substeps
    self.collide_scale = config.solver_scale_collide
    # updates only applied if velocity differences exceed this threshold
    self.velocity_threshold = jp.norm(vec_to_arr(config.gravity)) * self.h * 4.0

  def apply(self, qp: QP) -> Tuple[P, geometry.Contact]:
    """Returns impulse from any potential contacts between collidables.

    Args:
      qp: Coordinate/velocity frame of the bodies.

    Returns:
      dP: Impulse to apply to the bodies in the collision.
    """
    col_a, col_b = self.cull.get()
    qp_a = jp.take(qp, col_a.body.idx)
    qp_b = jp.take(qp, col_b.body.idx)
    contact = jp.vmap(self.contact_fn)(col_a, col_b, qp_a, qp_b)
    pre_contact = contact
    dp_a, dp_b = jp.vmap(self._contact)(col_a, col_b, qp_a, qp_b, contact)

    rep_a = dp_a.vel.shape[1]
    rep_b = dp_b.vel.shape[1] if dp_b else None
    dp_a, dp_b = jp.tree_map(jp.concatenate, (dp_a, dp_b))

    if dp_b is None:
      dp_vel, dp_ang, body_idx = dp_a.vel, dp_a.ang, jp.repeat(
          col_a.body.idx, rep_a)
    else:
      body_idx = jp.concatenate(
          (jp.repeat(col_a.body.idx, rep_a), jp.repeat(col_b.body.idx, rep_b)))
      dp_vel = jp.concatenate((dp_a.vel, dp_b.vel))
      dp_ang = jp.concatenate((dp_a.ang, dp_b.ang))
    contact = jp.where(jp.any(dp_vel, axis=-1), 1.0, 0.0)
    contact = jp.segment_sum(contact, body_idx, qp.pos.shape[0])
    dp_vel = jp.segment_sum(dp_vel, body_idx, qp.pos.shape[0])
    dp_ang = jp.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    # equally distribute impulse over possible contacts
    contact = jp.reshape(1e-8 + contact, (dp_vel.shape[0], 1))
    dp_vel = dp_vel / contact
    dp_ang = dp_ang / contact
    return P(vel=dp_vel, ang=dp_ang), pre_contact

  def velocity_apply(self, qp: QP, dlambda: jp.ndarray, qp_prev: QP,
                     contact) -> P:
    """Returns impulse from any potential contacts between collidables.

    Args:
      qp: Coordinate/velocity frame of the bodies.
      dlambda: Normal forces from position based collision pass
      qp_prev: State data before the collision pass
      contact: Contact data from the last collision pass

    Returns:
      dP: Impulse to apply to the bodies in the collision.
    """
    col_a, col_b = self.cull.get()
    qp_a, qp_a_prev = jp.take((qp, qp_prev), col_a.body.idx)
    qp_b, qp_b_prev = jp.take((qp, qp_prev), col_b.body.idx)
    dp_a, dp_b = jp.vmap(self._velocity_contact)(col_a, col_b, qp_a, qp_b,  # pytype: disable=attribute-error
                                                 contact, dlambda, qp_a_prev,
                                                 qp_b_prev)

    rep_a = dp_a.vel.shape[1]
    rep_b = dp_b.vel.shape[1] if dp_b else None
    dp_a, dp_b = jp.tree_map(jp.concatenate, (dp_a, dp_b))

    if dp_b is None:
      dp_vel, dp_ang, body_idx = dp_a.vel, dp_a.ang, jp.repeat(
          col_a.body.idx, rep_a)
    else:
      body_idx = jp.concatenate(
          (jp.repeat(col_a.body.idx, rep_a), jp.repeat(col_b.body.idx, rep_b)))
      dp_vel = jp.concatenate((dp_a.vel, dp_b.vel))
      dp_ang = jp.concatenate((dp_a.ang, dp_b.ang))
    contact = jp.where(jp.any(dp_vel, axis=-1), 1.0, 0.0)
    contact = jp.segment_sum(contact, body_idx, qp.pos.shape[0])
    dp_vel = jp.segment_sum(dp_vel, body_idx, qp.pos.shape[0])
    dp_ang = jp.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    # equally distribute impulse over possible contacts
    contact = jp.reshape(1e-6 + contact, (dp_vel.shape[0], 1))
    dp_vel = dp_vel / contact
    dp_ang = dp_ang / contact
    return P(vel=dp_vel, ang=dp_ang)

  def position_apply(self, qp: QP,
                     qp_prev: QP) -> Tuple[Q, jp.ndarray, geometry.Contact]:
    """Returns a position based update that resolves a collisions for QP.

    Args:
      qp: Most recent state data for the system.
      qp_prev: State data before the most recent kinetic update.

    Returns:
      dQ: Changes in position and quaternion to enforce collision.
    """
    col_a, col_b = self.cull.get()
    qp_a, qp_a_prev = jp.take((qp, qp_prev), col_a.body.idx)
    qp_b, qp_b_prev = jp.take((qp, qp_prev), col_b.body.idx)

    contact = jp.vmap(self.contact_fn)(col_a, col_b, qp_a, qp_b)
    pre_contact = contact
    dq_a, dq_b, dlambda = jp.vmap(self._position_contact)(col_a, col_b, qp_a,
                                                          qp_b, qp_a_prev,
                                                          qp_b_prev, contact)

    rep_a = dq_a.pos.shape[1]
    rep_b = dq_b.pos.shape[1] if dq_b else None
    dq_a, dq_b = jp.tree_map(jp.concatenate, (dq_a, dq_b))

    if dq_b is None:
      dq_pos, dq_rot, body_idx = dq_a.pos, dq_a.rot, jp.repeat(
          col_a.body.idx, rep_a)
    else:
      body_idx = jp.concatenate(
          (jp.repeat(col_a.body.idx, rep_b), jp.repeat(col_b.body.idx, rep_b)))
      dq_pos = jp.concatenate((dq_a.pos, dq_b.pos))
      dq_rot = jp.concatenate((dq_a.rot, dq_b.rot))
    contact = jp.where(jp.any(dq_pos, axis=-1), 1.0, 0.0)
    contact = jp.segment_sum(contact, body_idx, qp.pos.shape[0])
    dq_pos = jp.segment_sum(dq_pos, body_idx, qp.pos.shape[0])
    dq_rot = jp.segment_sum(dq_rot, body_idx, qp.rot.shape[0])

    # equally distribute impulse over possible contacts
    contact = jp.reshape(1e-6 + contact, (dq_pos.shape[0], 1))
    dq_pos = dq_pos / contact
    dq_rot = dq_rot / contact
    return Q(pos=dq_pos, rot=dq_rot), dlambda, pre_contact

  @abc.abstractmethod
  def _contact(self, col_a: geometry.Collidable, col_b: geometry.Collidable,
               qp_a: QP, qp_b: QP,
               contact: geometry.Contact) -> Tuple[P, Optional[P]]:
    pass

  @abc.abstractmethod
  def _position_contact(
      self, col_a: geometry.Collidable, col_b: geometry.Collidable, qp_a: QP,
      qp_b: QP, qp_a_old: QP, qp_b_old: QP,
      contact: geometry.Contact) -> Tuple[Q, Optional[Q], jp.ndarray]:
    pass

  @abc.abstractmethod
  def _velocity_contact(self, col_a: geometry.Collidable,
                        col_b: geometry.Collidable, qp_a: QP, qp_b: QP,
                        contact: geometry.Contact, dlambda: jp.ndarray,
                        qp_a_old: QP, qp_b_old: QP) -> Tuple[P, Optional[P]]:
    pass


@pytree.register
class OneWayCollider(Collider):
  """Calculates one-way impulses, where the second collidable is static."""

  def _contact(self, col_a: geometry.Collidable, col_b: geometry.Collidable,
               qp_a: QP, qp_b: QP,
               contact: geometry.Contact) -> Tuple[P, Optional[P]]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact):
      # there are a few ways to combine material properties during contact.
      # multiplying is a reasonable default.  in the future we may allow others
      elasticity = col_a.elasticity * col_b.elasticity
      friction = col_a.friction * col_b.friction
      rel_pos = contact.pos - qp_a.pos
      baumgarte_vel = self.baumgarte_erp * contact.penetration
      normal_vel = jp.dot(contact.normal, contact.vel)
      temp1 = col_a.body.inertia * jp.cross(rel_pos, contact.normal)
      ang = jp.dot(contact.normal, jp.cross(temp1, rel_pos))
      impulse = (-1. * (1. + elasticity) * normal_vel + baumgarte_vel) / (
          (1. / col_a.body.mass) + ang)
      dp_n = col_a.body.impulse(qp_a, impulse * contact.normal, contact.pos)

      # apply drag due to friction acting parallel to the surface contact
      vel_d = contact.vel - normal_vel * contact.normal
      impulse_d = jp.safe_norm(vel_d) / ((1. / (col_a.body.mass)) + ang)
      # drag magnitude cannot exceed max friction
      impulse_d = jp.minimum(impulse_d, friction * impulse)
      dir_d = vel_d / (1e-6 + jp.safe_norm(vel_d))
      dp_d = col_a.body.impulse(qp_a, -impulse_d * dir_d, contact.pos)
      # apply collision if penetrating, approaching, and oriented correctly
      apply_n = jp.where(
          (contact.penetration > 0.) & (normal_vel < 0) & (impulse > 0.), 1.,
          0.)
      # apply drag if moving laterally above threshold
      apply_d = apply_n * jp.where(jp.safe_norm(vel_d) > 0.01, 1., 0.)

      dp_a = dp_n * apply_n + dp_d * apply_d
      return dp_a, None

    return _v_contact(contact)

  def _position_contact(
      self, col_a: geometry.Collidable, col_b: geometry.Collidable, qp_a: QP,
      qp_b: QP, qp_a_old: QP, qp_b_old: QP,
      contact: geometry.Contact) -> Tuple[Q, Optional[Q], jp.ndarray]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact):

      friction = col_a.friction * col_b.friction

      pos_p = contact.pos
      pos_c = contact.pos + contact.normal * contact.penetration
      dx = pos_p - pos_c
      pos_p = pos_p - qp_a.pos
      pos_c = pos_c - qp_b.pos

      n = contact.normal
      c = jp.dot(dx, n)

      # only spherical inertias for now
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      dlambda = -c / (w1 + 1e-6)

      coll_mask = jp.where(c < 0, 1., 0.)
      p = dlambda * n * coll_mask

      dq_p_pos = p / col_a.body.mass
      dq_p_rot = .5 * math.vec_quat_mul(col_a.body.inertia * jp.cross(pos_p, p),
                                        qp_a.rot)

      dq_p = Q(
          pos=self.collide_scale * dq_p_pos, rot=self.collide_scale * dq_p_rot)

      # static friction

      q1inv = math.quat_inv(qp_a.rot)
      r1 = math.rotate(contact.pos - qp_a.pos, q1inv)

      p1bar = qp_a_old.pos + math.rotate(r1, qp_a_old.rot)
      p1 = contact.pos

      deltap = p1 - p1bar
      deltap_t = deltap - jp.dot(deltap, n) * n

      dx = deltap_t

      c = jp.safe_norm(dx)
      n = dx / (c + 1e-6)

      # using spherical inertia
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      dlambdat = -c / (w1 + 0.)
      static_mask = jp.where(
          jp.abs(dlambdat) < jp.abs(friction * dlambda), 1., 0.)
      p = dlambdat * n * static_mask * coll_mask

      dq_p_pos = p / col_a.body.mass
      dq_p_rot = .5 * math.vec_quat_mul(col_a.body.inertia * jp.cross(pos_p, p),
                                        qp_a.rot)

      dq_p = Q(
          pos=dq_p.pos + self.collide_scale * dq_p_pos,
          rot=dq_p.rot + self.collide_scale * dq_p_rot)

      return dq_p, None, dlambda * coll_mask

    return _v_contact(contact)

  def _velocity_contact(self, col_a: geometry.Collidable,
                        col_b: geometry.Collidable, qp_a: QP, qp_b: QP,
                        contact: geometry.Contact, dlambda: jp.ndarray,
                        qp_a_old: QP, qp_b_old: QP) -> Tuple[P, Optional[P]]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact, dlambda):
      # there are a few ways to combine material properties during contact.
      # multiplying is a reasonable default.  in the future we may allow others

      # dynamic friction calculation
      elasticity = col_a.elasticity * col_b.elasticity
      friction = col_a.friction * col_b.friction

      n = contact.normal
      rel_vel = qp_a.vel + jp.cross(qp_a.ang, contact.pos - qp_a.pos)
      v_n = jp.dot(rel_vel, n)
      v_t = rel_vel - n * v_n
      v_t_norm = jp.safe_norm(v_t)
      v_t_dir = v_t / (1e-6 + v_t_norm)

      # factor of 2 from integrator doing 1 collision pass every 2 steps
      dvel = -v_t_dir * jp.amin(
          jp.array([friction * jp.abs(dlambda) / (2. * self.h), v_t_norm]))

      angw = jp.cross((contact.pos - qp_a.pos), v_t_dir)
      w = (1. / col_a.body.mass) + jp.dot(angw, angw)

      p_dyn = dvel / (w + 1e-6)

      # restitution calculation

      rel_vel_old = qp_a_old.vel + jp.cross(qp_a_old.ang,
                                            contact.pos - qp_a_old.pos)
      v_n_old = jp.dot(rel_vel_old, n)

      dv_rest = n * (-v_n - jp.amin(jp.array([elasticity * v_n_old, 0.])))

      pos_p = contact.pos

      dx = dv_rest
      pos_p = pos_p - qp_a.pos

      c = jp.safe_norm(dx)
      n = dx / (c + 1e-6)

      # only spherical inertia effects
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      dlambda_rest = c / (w1 + 1e-6)
      static_mask = jp.where(contact.penetration > 0, 1., 0.)
      sinking = jp.where(v_n_old <= -self.velocity_threshold, 1., 0.)

      p = (dlambda_rest * n * sinking + p_dyn) * static_mask

      dp_p = P(
          vel=p / col_a.body.mass,
          ang=jp.cross(col_a.body.inertia * (contact.pos - qp_a.pos), p))

      return dp_p, None

    return _v_contact(contact, dlambda)


@pytree.register
class TwoWayCollider(Collider):
  """Calculates two-way impulses on collidable pairs."""

  def _contact(self, col_a: geometry.Collidable, col_b: geometry.Collidable,
               qp_a: QP, qp_b: QP,
               contact: geometry.Contact) -> Tuple[P, Optional[P]]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact):
      # there are a few ways to combine material properties during contact.
      # multiplying is a reasonable default.  in the future we may allow others
      elasticity = col_a.elasticity * col_b.elasticity
      friction = col_a.friction * col_b.friction
      rel_pos_a = contact.pos - qp_a.pos
      rel_pos_b = contact.pos - qp_b.pos
      baumgarte_vel = self.baumgarte_erp * contact.penetration
      normal_vel = jp.dot(contact.normal, contact.vel)
      temp1 = col_a.body.inertia * jp.cross(rel_pos_a, contact.normal)
      temp2 = col_b.body.inertia * jp.cross(rel_pos_b, contact.normal)
      ang = jp.dot(contact.normal,
                   jp.cross(temp1, rel_pos_a) + jp.cross(temp2, rel_pos_b))
      impulse = (-1. * (1. + elasticity) * normal_vel + baumgarte_vel) / (
          (1. / col_a.body.mass) + (1. / col_b.body.mass) + ang)
      dp_n_a = col_a.body.impulse(qp_a, impulse * contact.normal, contact.pos)
      dp_n_b = col_b.body.impulse(qp_b, -impulse * contact.normal, contact.pos)

      # apply drag due to friction acting parallel to the surface contact
      vel_d = contact.vel - normal_vel * contact.normal
      impulse_d = jp.safe_norm(vel_d) / ((1. / col_a.body.mass) +
                                         (1. / col_b.body.mass) + ang)
      # drag magnitude cannot exceed max friction
      impulse_d = jp.minimum(impulse_d, friction * impulse)
      dir_d = vel_d / (1e-6 + jp.safe_norm(vel_d))
      dp_d_a = col_a.body.impulse(qp_a, -impulse_d * dir_d, contact.pos)
      dp_d_b = col_b.body.impulse(qp_b, impulse_d * dir_d, contact.pos)
      # apply collision if penetrating, approaching, and oriented correctly
      apply_n = jp.where(
          (contact.penetration > 0.) & (normal_vel < 0) & (impulse > 0.), 1.,
          0.)
      # apply drag if moving laterally above threshold
      apply_d = apply_n * jp.where(jp.safe_norm(vel_d) > 0.01, 1., 0.)

      dp_a = dp_n_a * apply_n + dp_d_a * apply_d
      dp_b = dp_n_b * apply_n + dp_d_b * apply_d
      return dp_a, dp_b

    return _v_contact(contact)

  def _position_contact(
      self, col_a: geometry.Collidable, col_b: geometry.Collidable, qp_a: QP,
      qp_b: QP, qp_a_old: QP, qp_b_old: QP,
      contact: geometry.Contact) -> Tuple[Q, Optional[Q], jp.ndarray]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact):
      pos_p = contact.pos - contact.normal * contact.penetration / 2.
      pos_c = contact.pos + contact.normal * contact.penetration / 2.
      pos_p = pos_p - qp_a.pos
      pos_c = pos_c - qp_b.pos

      n = contact.normal
      c = -contact.penetration

      # only spherical inertia effects
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      cr2 = jp.cross(pos_c, n)
      w2 = (1. / col_b.body.mass) + jp.dot(cr2, col_b.body.inertia * cr2)

      dlambda = -c / (w1 + w2 + 1e-6)
      coll_mask = jp.where(c < 0, 1., 0.)
      p = dlambda * n * coll_mask

      dq_p_pos = p / col_a.body.mass
      dq_p_rot = .5 * math.vec_quat_mul(col_a.body.inertia * jp.cross(pos_p, p),
                                        qp_a.rot)

      dq_c_pos = -p / col_b.body.mass
      dq_c_rot = -.5 * math.vec_quat_mul(
          col_b.body.inertia * jp.cross(pos_c, p), qp_b.rot)

      dq_p = Q(
          pos=self.collide_scale * dq_p_pos, rot=self.collide_scale * dq_p_rot)
      dq_c = Q(
          pos=self.collide_scale * dq_c_pos, rot=self.collide_scale * dq_c_rot)

      # static friction stuff

      q1inv = math.quat_inv(qp_a.rot)
      r1 = math.rotate(contact.pos - qp_a.pos, q1inv)

      q2inv = math.quat_inv(qp_b.rot)
      r2 = math.rotate(contact.pos - qp_b.pos, q2inv)

      p1bar = qp_a_old.pos + math.rotate(r1, qp_a_old.rot)
      p2bar = qp_b_old.pos + math.rotate(r2, qp_b_old.rot)
      p0 = contact.pos

      deltap = (p0 - p1bar) - (p0 - p2bar)
      deltap_t = deltap - jp.dot(deltap, n) * n

      pos_p = contact.pos - qp_a.pos
      pos_c = contact.pos - qp_b.pos

      c = jp.safe_norm(deltap_t)
      n = deltap_t / (c + 1e-6)

      # ignoring inertial effects for now
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      cr2 = jp.cross(pos_c, n)
      w2 = (1. / col_b.body.mass) + jp.dot(cr2, col_b.body.inertia * cr2)

      dlambdat = -c / (w1 + w2)
      static_mask = jp.where(jp.abs(dlambdat) < jp.abs(dlambda), 1., 0.)
      p = dlambdat * n * static_mask * coll_mask

      dq_p_pos = p / col_a.body.mass
      dq_p_rot = .5 * math.vec_quat_mul(col_a.body.inertia * jp.cross(pos_p, p),
                                        qp_a.rot)

      dq_c_pos = -p / col_b.body.mass
      dq_c_rot = .5 * math.vec_quat_mul(
          col_b.body.inertia * jp.cross(pos_c, -p), qp_b.rot)

      dq_p += Q(
          pos=self.collide_scale * dq_p_pos, rot=self.collide_scale * dq_p_rot)
      dq_c += Q(
          pos=self.collide_scale * dq_c_pos, rot=self.collide_scale * dq_c_rot)

      return dq_p, dq_c, dlambda  # pytype: disable=bad-return-type

    return _v_contact(contact)  # pytype: disable=bad-return-type

  def _velocity_contact(self, col_a: geometry.Collidable,
                        col_b: geometry.Collidable, qp_a: QP, qp_b: QP,
                        contact: geometry.Contact, dlambda: jp.ndarray,
                        qp_a_old: QP, qp_b_old: QP) -> Tuple[P, Optional[P]]:
    """Calculates impulse on a body due to a contact."""

    @jp.vmap
    def _v_contact(contact, dlambda):
      # there are a few ways to combine material properties during contact.
      # multiplying is a reasonable default.  in the future we may allow others

      # dynamic friction calculation
      friction = col_a.friction * col_b.friction
      elasticity = col_a.elasticity * col_b.elasticity
      n = contact.normal
      rel_vel = qp_a.vel + jp.cross(qp_a.ang, contact.pos - qp_a.pos) - (
          qp_b.vel + jp.cross(qp_b.ang, contact.pos - qp_b.pos))
      v_n = jp.dot(rel_vel, n)
      v_t = rel_vel - n * v_n
      v_t_norm = jp.safe_norm(v_t)
      v_t_dir = v_t / (1e-6 + v_t_norm)

      # factor of 2 from integrator doing 1 collision pass every 2 steps
      dvel = -v_t_dir * jp.amin(
          jp.array([friction * jp.abs(dlambda) / (2. * self.h), v_t_norm]))

      angw_1 = jp.cross((contact.pos - qp_a.pos), v_t_dir)
      angw_2 = jp.cross((contact.pos - qp_b.pos), v_t_dir)
      w1 = (1. / col_a.body.mass) + jp.dot(angw_1, col_a.body.inertia * angw_1)
      w2 = (1. / col_b.body.mass) + jp.dot(angw_2, col_b.body.inertia * angw_2)

      p_dyn = dvel / (w1 + w2 + 1e-6)

      # restitution calculation

      rel_vel_old = (
          qp_a_old.vel + jp.cross(qp_a_old.ang, contact.pos - qp_a_old.pos)) - (
              qp_b_old.vel + jp.cross(qp_b_old.ang, contact.pos - qp_b_old.pos))
      v_n_old = jp.dot(rel_vel_old, n)

      dv_rest = n * (-v_n - jp.amin(jp.array([elasticity * v_n_old, 0.])))

      pos_p = contact.pos
      pos_c = contact.pos + contact.normal * contact.penetration
      dx = dv_rest
      pos_p = pos_p - qp_a.pos
      pos_c = pos_c - qp_b.pos

      c = jp.safe_norm(dx)
      n = dx / (c + 1e-6)

      # ignoring inertial effects for now
      cr1 = jp.cross(pos_p, n)
      w1 = (1. / col_a.body.mass) + jp.dot(cr1, col_a.body.inertia * cr1)

      cr2 = jp.cross(pos_c, n)
      w2 = (1. / col_b.body.mass) + jp.dot(cr2, col_b.body.inertia * cr2)

      dlambda_rest = c / (w1 + w2 + 1e-6)
      static_mask = jp.where(contact.penetration > 0, 1., 0.)
      sinking = jp.where(v_n_old <= 0., 1., 0.)

      p = (dlambda_rest * n * sinking + p_dyn) * static_mask

      dp_p = P(
          vel=p / col_a.body.mass,
          ang=jp.cross(col_a.body.inertia * (contact.pos - qp_a.pos), p))

      dp_c = P(
          vel=-p / col_b.body.mass,
          ang=jp.cross(col_b.body.inertia * (contact.pos - qp_b.pos), -p))

      return dp_p, dp_c

    return _v_contact(contact, dlambda)


def _endpoints(end: jp.ndarray, qp: QP, offset: jp.ndarray):
  pos = qp.pos + math.rotate(offset, qp.rot)
  end = math.rotate(end, qp.rot)
  return pos + end, pos - end


def box_plane(box: geometry.Box, _: geometry.Plane, qp_a: QP,
              qp_b: QP) -> geometry.Contact:
  """Returns vectorized contacts between a box and a plane."""

  @jp.vmap
  def point_plane(corner):
    pos, vel = qp_a.to_world(corner)
    normal = math.rotate(jp.array([0., 0., 1.]), qp_b.rot)
    penetration = jp.dot(qp_b.pos - pos, normal)
    return pos, vel, normal, penetration

  pos, vel, normal, penetration = point_plane(box.corner)

  return geometry.Contact(pos, vel, normal, penetration)


def mesh_plane(mesh: geometry.Mesh, _: geometry.Plane, qp_a: QP,
               qp_b: QP) -> geometry.Contact:
  """Similar to box-plane collision, but uses vertices instead of corners."""

  @jp.vmap
  def point_plane(vertices):
    pos, vel = qp_a.to_world(vertices)
    normal = math.rotate(jp.array([0., 0., 1.]), qp_b.rot)
    penetration = jp.dot(qp_b.pos - pos, normal)
    return pos, vel, normal, penetration

  pos, vel, normal, penetration = point_plane(mesh.vertices)

  return geometry.Contact(pos, vel, normal, penetration)


def box_heightmap(box: geometry.Box, hm: geometry.HeightMap, qp_a: QP,
                  qp_b: QP) -> geometry.Contact:
  """Returns contact between a box corner and a height map."""
  # TODO: this only checks box corners against height map surfaces,
  # and is missing box planes against height map points.
  # TODO: collisions are not well defined outside of the height
  # map coordinates.
  @jp.vmap
  def corner_heightmap(corner):
    box_pos, vel = qp_a.to_world(corner)
    pos = math.inv_rotate(box_pos - qp_b.pos, qp_b.rot)
    uv_pos = pos[:2] / hm.cell_size

    # Find the square mesh that enclose the candidate point, which we split into
    # two triangles.
    uv_idx = jp.floor(uv_pos).astype(int)
    delta_uv = uv_pos - uv_idx
    lower_triangle = jp.sum(delta_uv) < 1
    mu = jp.where(lower_triangle, -1, 1)

    # Compute the triangle vertices (u, v) that enclose the candidate point.
    triangle_u = uv_idx[0] + jp.where(lower_triangle, jp.array([0, 1, 0]),
                                      jp.array([1, 0, 1]))
    triangle_v = uv_idx[1] + jp.where(lower_triangle, jp.array([0, 0, 1]),
                                      jp.array([1, 1, 0]))

    # Get the heights for each triangle vertice. Height data is stored in row
    # major order where the row index is the x-position and the column index is
    # the y-position. The heightmap origin (u, v) is at the top-left corner.
    h = hm.height[triangle_u, -triangle_v]

    raw_normal = jp.array(
        [mu * (h[1] - h[0]), mu * (h[2] - h[0]), hm.cell_size])
    normal = raw_normal / jp.safe_norm(raw_normal)
    p0 = jp.array(
        [triangle_u[0] * hm.cell_size, triangle_v[0] * hm.cell_size, h[0]])
    penetration = jp.dot(p0 - pos, normal)
    # Rotate back to the world frame.
    normal = math.rotate(normal, qp_b.rot)
    return box_pos, vel, normal, penetration

  pos, vel, normal, penetration = corner_heightmap(box.corner)
  return geometry.Contact(pos, vel, normal, penetration)


def capsule_plane(cap: geometry.CapsuleEnd, _: geometry.Plane, qp_a: QP,
                  qp_b: QP) -> geometry.Contact:
  """Returns contact between a capsule and a plane."""

  @jp.vmap
  def sphere_plane(end):
    cap_end_world = qp_a.pos + math.rotate(end, qp_a.rot)
    normal = math.rotate(jp.array([0., 0., 1.]), qp_b.rot)
    pos = cap_end_world - normal * cap.radius
    vel = qp_a.vel + jp.cross(qp_a.ang, pos - qp_a.pos)
    penetration = jp.dot(qp_b.pos - pos, normal)
    return pos, vel, normal, penetration

  pos, vel, normal, penetration = sphere_plane(cap.end)

  return geometry.Contact(pos, vel, normal, penetration)


def capsule_clippedplane(cap: geometry.CapsuleEnd, plane: geometry.ClippedPlane,
                         qp_a: QP, qp_b: QP) -> geometry.Contact:
  """Returns contact between a capsule and a clipped plane."""

  @jp.vmap
  def sphere_clippedplane(end):
    cap_end_world = qp_a.pos + math.rotate(end, qp_a.rot)
    normal = math.rotate(plane.normal, qp_b.rot)

    # orient the normal s.t. it points at the CoM of the capsule
    normal_dir = jp.where(qp_a.pos.dot(normal) > 0., 1, -1)
    normal = normal * normal_dir

    pos = cap_end_world - normal * cap.radius
    vel = qp_a.vel + jp.cross(qp_a.ang, pos - qp_a.pos)
    plane_pt = math.rotate(plane.pos, qp_b.rot) + qp_b.pos
    penetration = jp.dot(plane_pt - pos, normal)

    # Clip against side planes.
    norm_x = math.rotate(plane.x, qp_b.rot)
    norm_y = math.rotate(plane.y, qp_b.rot)
    side_plane_pt = jp.array([
        plane_pt + norm_x * plane.halfsize_x,
        plane_pt - norm_x * plane.halfsize_x,
        plane_pt + norm_y * plane.halfsize_y,
        plane_pt - norm_y * plane.halfsize_y])
    yn, xn = jp.cross(normal * normal_dir, norm_x), -jp.cross(
        normal * normal_dir, norm_y)
    side_plane_norm = jp.array([xn, -xn, yn, -yn])
    in_front_of_side_plane = jp.vmap(geometry.point_in_front_of_plane,
                                     include=[True, True, False])(
                                         side_plane_pt, side_plane_norm, pos)
    penetration = jp.where(jp.any(in_front_of_side_plane),
                           -jp.ones_like(penetration),
                           penetration)

    return pos, vel, normal, penetration

  pos, vel, normal, penetration = sphere_clippedplane(cap.end)

  return geometry.Contact(pos, vel, normal, penetration)


def capsule_capsule(cap_a: geometry.Capsule, cap_b: geometry.Capsule, qp_a: QP,
                    qp_b: QP) -> geometry.Contact:
  """Returns contact between two capsules."""
  a0, a1 = _endpoints(cap_a.end, qp_a, cap_a.pos)
  b0, b1 = _endpoints(cap_b.end, qp_b, cap_b.pos)
  a_best, b_best = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)

  penetration_vec = a_best - b_best
  dist = jp.safe_norm(penetration_vec)
  normal = penetration_vec / (1e-6 + dist)
  penetration = cap_a.radius + cap_b.radius - dist
  pos = (a_best + b_best) / 2
  vel = qp_a.world_velocity(pos) - qp_b.world_velocity(pos)
  out = geometry.Contact(pos, vel, normal, penetration)
  return jp.tree_map(jp.expand_dims, out)


def capsule_mesh(cap: geometry.Capsule, mesh: geometry.BaseMesh, qp_a: QP,
                 qp_b: QP) -> geometry.Contact:
  """Returns the contacts for capsule-mesh collision."""

  @jp.vmap
  def capsule_face(faces, face_normals):
    # Determine the capsule line.
    a, b = _endpoints(cap.end, qp_a, cap.pos)
    triangle_normal = math.rotate(face_normals, qp_b.rot)

    pt = qp_b.pos + jp.vmap(math.rotate, include=[True, False])(faces, qp_b.rot)
    p0, p1, p2 = pt[..., 0, :], pt[..., 1, :], pt[..., 2, :]

    segment_p, triangle_p = geometry.closest_segment_triangle_points(
        a, b, p0, p1, p2, triangle_normal)

    penetration_vec = segment_p - triangle_p
    dist = jp.safe_norm(penetration_vec)
    normal = penetration_vec / (1e-6 + dist)
    penetration = cap.radius - dist

    pos = triangle_p
    vel = qp_a.world_velocity(pos) - qp_b.world_velocity(pos)
    return pos, vel, normal, penetration

  pos, vel, normal, penetration = capsule_face(mesh.faces, mesh.face_normals)
  return geometry.Contact(pos, vel, normal, penetration)


def hull_hull(mesh_a: geometry.BaseMesh, mesh_b: geometry.BaseMesh, qp_a: QP,
              qp_b: QP) -> geometry.Contact:
  """Gets hull-hull contacts."""

  @jp.vmap
  def get_faces(faces_a, faces_b, normals_a, normals_b):
    faces_a = qp_a.pos + jp.vmap(
        math.rotate, include=[True, False])(faces_a, qp_a.rot)
    faces_b = qp_b.pos + jp.vmap(
        math.rotate, include=[True, False])(faces_b, qp_b.rot)
    normals_a = math.rotate(normals_a, qp_a.rot)
    normals_b = math.rotate(normals_b, qp_b.rot)
    return faces_a, faces_b, normals_a, normals_b

  @jp.vmap
  def get_verts(vertices_a, vertices_b):
    vertices_a = qp_a.pos + math.rotate(vertices_a, qp_a.rot)
    vertices_b = qp_b.pos + math.rotate(vertices_b, qp_b.rot)
    return vertices_a, vertices_b

  faces_a, faces_b, normals_a, normals_b = get_faces(mesh_a.faces, mesh_b.faces,
                                                     mesh_a.face_normals,
                                                     mesh_b.face_normals)
  vertices_a, vertices_b = get_verts(mesh_a.vertices, mesh_b.vertices)

  # Create a potential face and edge contact using SAT.
  edge_contact, face_contact = geometry.sat_hull_hull(faces_a, faces_b,
                                                      vertices_a, vertices_b,
                                                      normals_a, normals_b)

  # Pick a face or edge as the final contact.
  contact = jp.cond(edge_contact.penetration[0] > 0, lambda *x: edge_contact,
                    lambda *x: face_contact)

  get_vel = lambda p: qp_a.world_velocity(p) - qp_b.world_velocity(p)
  contact.vel = jp.vmap(get_vel)(contact.pos)

  return contact


def get(config: config_pb2.Config, body: bodies.Body) -> List[Collider]:
  """Creates all colliders given a config."""

  mesh_geoms = {mg.name: mg for mg in config.mesh_geometries}
  collider_pairs = {
      ('box', 'plane'): (geometry.Box, geometry.Plane, box_plane),
      ('box', 'heightMap'): (geometry.Box, geometry.HeightMap, box_heightmap),
      ('capsule', 'box'):
          (geometry.Capsule, geometry.TriangulatedBox, capsule_mesh),
      ('capsule', 'plane'):
          (geometry.CapsuleEnd, geometry.Plane, capsule_plane),
      ('capsule', 'capsule'):
          (geometry.Capsule, geometry.Capsule, capsule_capsule),
      ('capsule', 'mesh'):
          (geometry.Capsule,
           functools.partial(geometry.Mesh,
                             mesh_geoms=mesh_geoms), capsule_mesh),
      ('capsule', 'clipped_plane'): (geometry.CapsuleEnd, geometry.ClippedPlane,
                                     capsule_clippedplane),
      ('mesh', 'plane'):
          (functools.partial(geometry.PointMesh, mesh_geoms=mesh_geoms),
           geometry.Plane, mesh_plane),
      ('box', 'box'): (geometry.HullBox, geometry.HullBox, hull_hull),
  }
  supported_near_neighbors = {('capsule', 'capsule')}
  unique_meshes = {}
  cols = []
  for b in config.bodies:
    for c_idx, c in enumerate(b.colliders):
      if c.no_contact:
        continue
      if c.WhichOneof('type') == 'sphere':
        c = c if isinstance(c, config_pb2.Collider) else c.msg
        nc = config_pb2.Collider()
        nc.CopyFrom(c)
        nc.capsule.radius = c.sphere.radius
        nc.capsule.length = 2 * c.sphere.radius
        nc.capsule.end = 1
        c = nc
      if c.WhichOneof('type') == 'mesh':
        unique_meshes[c.mesh.name] = 1
      cols.append((c, b, c_idx))

  include = {(ci.first, ci.second) for ci in config.collide_include}
  parents = ((j.parent, j.child) for j in config.joints)
  ret = []

  for (type_a, type_b), (cls_a, cls_b, contact_fn) in collider_pairs.items():
    replicas = unique_meshes.keys() if 'mesh' in (type_a, type_b) else [None]
    for mesh_name in replicas:
      cols_a, cols_b = [], []
      for (cols_i, type_i) in [(cols_a, type_a), (cols_b, type_b)]:
        for c, b, c_idx in cols:
          if c.WhichOneof('type') == type_i:
            if type_i == 'mesh' and c.mesh.name != mesh_name:
              continue  # only add colliders with the same mesh
            cols_i.append((c, b, c_idx))

      cols_a = [(c, b, c_idx) for c, b, c_idx in cols_a if not b.frozen.all]

      cols_ab = []

      pair_count = {}
      for ca, ba, ca_idx in cols_a:
        for cb, bb, cb_idx in cols_b:
          included = (ba.name, bb.name) in include or (bb.name,
                                                       ba.name) in include
          if (ba.name, ca_idx, bb.name,
              cb_idx) in pair_count or (bb.name, cb_idx, ba.name,
                                        ca_idx) in pair_count:
            continue  # don't double count collisions
          if ba.name == bb.name:
            continue  # no self-collision
          if ba.frozen.all and bb.frozen.all:
            continue  # two immovable bodies cannot collide
          if (ba.name, bb.name
             ) in parents or (bb.name, ba.name) in parents and not included:
            continue  # ignore colliders for joint parents and children
          if ca.no_contact or cb.no_contact:
            continue  # ignore colliders that are purely visual

          if not include or included:
            # two cases:
            # 1. includes in config. then, only include if set in config
            # 2. includes not set in config, then include all collisions
            cols_ab.append((ca, ca_idx, ba, cb, cb_idx, bb))
            pair_count[(ba.name, ca_idx, bb.name, cb_idx)] = 1
            pair_count[(bb.name, cb_idx, ba.name, ca_idx)] = 1

      for b_is_frozen in (True, False):
        cols_ab_filtered = [
            x for x in cols_ab if x[-1].frozen.all == b_is_frozen
        ]
        if not cols_ab_filtered:
          continue

        bodies_a, bodies_b = [], []
        unique_check, unique_bodies = {}, []
        for (ca, ca_idx, ba, cb, cb_idx, bb) in cols_ab_filtered:
          for c, c_idx, b, arr in [(ca, ca_idx, ba, bodies_a),
                                   (cb, cb_idx, bb, bodies_b)]:
            nb = config_pb2.Body()
            bb = b if isinstance(b, config_pb2.Body) else b.msg
            cc = c if isinstance(c, config_pb2.Collider) else c.msg
            nb.CopyFrom(bb)
            nb.ClearField('colliders')
            nb.colliders.append(cc)
            if not isinstance(b, config_pb2.Body):
              arr.append(customize.TracedConfig(nb, custom_tree=b.custom_tree))
            else:
              arr.append(nb)
            if (b.name, c_idx) not in unique_check:
              unique_bodies.append(nb)
              unique_check[(b.name, c_idx)] = 1
        if config.collider_cutoff and len(
            bodies_a) > config.collider_cutoff and (
                type_a, type_b) in supported_near_neighbors:
          # pytype: disable=wrong-arg-types
          col_a = cls_a(bodies_a, body)
          col_b = cls_b(bodies_b, body)
          cull = NearNeighbors(
              cls_a(unique_bodies, body), cls_b(unique_bodies, body),
              (col_a.body.idx, col_b.body.idx), config.collider_cutoff)
        else:
          # pytype: disable=wrong-arg-types
          cull = Pairs(cls_a(bodies_a, body), cls_b(bodies_b, body))
        if b_is_frozen:
          collider = OneWayCollider(contact_fn, cull, config)
        else:
          collider = TwoWayCollider(contact_fn, cull, config)
        ret.append(collider)

  return ret



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for constraint satisfaction."""
from typing import Optional, Tuple

from brax.v2 import math
from brax.v2 import scan
from brax.v2.base import Contact, Motion, System, Transform
import jax
from jax import numpy as jp
import jaxopt
from jaxopt.projection import projection_non_negative


def _pt_jac(
    sys: System,
    com: jp.ndarray,
    cdof: Motion,
    pos: jp.ndarray,
    link_idx: jp.ndarray,
) -> jp.ndarray:
  """Calculates the point jacobian.

  Args:
    sys: a brax system
    com: center of mass position
    cdof: dofs in com frame
    pos: position in world frame
    link_idx: index of link frame to transform point jacobian

  Returns:
    pt: point jacobian
  """
  # backward scan up tree: build the link mask corresponding to link_idx
  def mask_fn(mask_child, link):
    mask = link == link_idx
    if mask_child is not None:
      mask += mask_child
    return mask

  mask = scan.tree(sys, mask_fn, 'l', jp.arange(sys.num_links()), reverse=True)
  cdof = jax.vmap(lambda a, b: a * b)(cdof, jp.take(mask, sys.dof_link()))
  pt = Transform.create(pos=pos - com).vmap(in_axes=(None, 0)).do(cdof).vel

  return pt


def _imp_aref(pos: jp.ndarray, vel: jp.ndarray) -> jp.ndarray:
  """Calculates impedance and offset acceleration in constraint frame.

  Args:
    pos: position in constraint frame
    vel: velocity in constraint frame

  Returns:
    imp: constraint impedance
    Aref: offset acceleration in constraint frame
  """
  # this formulation corresponds to the parameterization described here:
  # https://mujoco.readthedocs.io/en/latest/modeling.html#solver-parameters
  # TODO: support custom solimp, solref
  timeconst, dampratio = 0.02, 1.0
  dmin, dmax, width, mid, power = 0.9, 0.95, 0.001, 0.5, 2.0

  imp_x = jp.abs(pos) / width
  imp_a = (1.0 / jp.power(mid, power - 1)) * jp.power(imp_x, power)
  imp_b = 1 - (1.0 / jp.power(1 - mid, power - 1)) * jp.power(1 - imp_x, power)
  imp_y = jp.where(imp_x < mid, imp_a, imp_b)
  imp = dmin + imp_y * (dmax - dmin)
  imp = jp.clip(imp, dmin, dmax)
  imp = jp.where(imp_x > 1.0, dmax, imp)

  b = 2 / (dmax * timeconst)
  k = 1 / (dmax * dmax * timeconst * timeconst * dampratio * dampratio)

  aref = -b * vel - k * imp * pos

  return imp, aref


def jac_limit(
    sys: System, q: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
  """Calculates the jacobian for angle limits in dof frame.

  Args:
    sys: the brax system
    q: joint angle vector

  Returns:
    jac: the angle limit jacobian
    pos: angle in constraint frame
    diag: approximate diagonal of A matrix
  """
  if sys.dof.limit is None:
    return jp.zeros((0, sys.qd_size())), jp.zeros((0,)), jp.zeros((0,))

  # determine q and qd indices for non-free joints
  q_idx, qd_idx = sys.q_idx('123'), sys.qd_idx('123')

  pos_min = q[q_idx] - sys.dof.limit[0][qd_idx]
  pos_max = sys.dof.limit[1][qd_idx] - q[q_idx]
  pos = jp.minimum(jp.minimum(pos_min, pos_max), 0)

  side = ((pos_min < pos_max) * 2 - 1) * (pos < 0)
  jac = jax.vmap(jp.multiply)(jp.eye(sys.qd_size())[qd_idx], side)
  diag = sys.dof.invweight[qd_idx] * (pos < 0)

  return jac, pos, diag


def jac_contact(
    sys: System, com: jp.ndarray, cdof: Motion, contact: Optional[Contact]
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
  """Calculates the jacobian for contact constraints.

  Args:
    sys: the brax system
    com: center of mass position
    cdof: dofs in com frame
    contact: contacts computed for link geometries

  Returns:
    jac: the contact jacobian
    pos: contact position in constraint frame
    diag: approximate diagonal of A matrix
  """
  if contact is None:
    return jp.zeros((0, sys.qd_size())), jp.zeros((0,)), jp.zeros((0,))

  def row_fn(contact):
    link_a, link_b = contact.link_idx
    a = _pt_jac(sys, com, cdof, contact.pos, link_a)
    b = _pt_jac(sys, com, cdof, contact.pos, link_b)
    diff = b - a

    # 4 pyramidal friction directions
    jac = []
    for d in math.orthogonals(contact.normal):
      for f in [-contact.friction, contact.friction]:
        jac.append(diff @ (d * f - contact.normal))

    jac = jp.stack(jac)
    pos = -jp.tile(contact.penetration, 4)
    t = sys.link.invweight[link_a] + sys.link.invweight[link_b] * (link_b > -1)
    diag = jp.tile(t + contact.friction * contact.friction * t, 4)
    diag = 2 * contact.friction * contact.friction * diag

    return jax.tree_map(
        lambda x: x * (contact.penetration > 0), (jac, pos, diag)
    )

  return jax.tree_map(jp.concatenate, jax.vmap(row_fn)(contact))


def jacobian(
    sys: System,
    q: jp.ndarray,
    com: jp.ndarray,
    cdof: Motion,
    contact: Optional[Contact],
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
  """Calculates the full constraint jacobian and constraint position.

  Args:
    sys: a brax system
    q: joint position vector
    com: center of mass position
    cdof: dofs in com frame
    contact: contacts computed for link geometries

  Returns:
    jac: the constraint jacobian
    pos: position in constraint frame
    diag: approximate diagonal of A matrix
  """
  jpds = jac_contact(sys, com, cdof, contact), jac_limit(sys, q)

  return jax.tree_map(lambda *x: jp.concatenate(x), *jpds)


def force(
    sys: System,
    qd: jp.ndarray,
    qf_smooth: jp.ndarray,
    mass_mx_inv: jp.ndarray,
    jac: jp.ndarray,
    pos: jp.ndarray,
    diag: jp.ndarray,
) -> jp.ndarray:
  """Calculates forces in constraint frame.

  Args:
    sys: a brax system
    qd: joint velocity vector
    qf_smooth: joint force vector for smooth dynamics
    mass_mx_inv: inverse mass matrix
    jac: the constraint jacobian
    pos: position in constraint frame
    diag: approximate diagonal of A matrix

  Returns:
    qf_constraint: delta joint velocity, already time-integrated
  """
  if jac.shape[0] == 0:
    return jp.zeros_like(qd)

  # calculate A matrix and b vector
  imp, aref = _imp_aref(pos, jac @ qd)
  a = jac @ mass_mx_inv @ jac.T
  a = a + jp.diag(diag * (1 - imp) / imp)
  b = jac @ mass_mx_inv @ qf_smooth - aref

  # solve for forces in constraint frame, Ax + b = 0 s.t. x >= 0
  def objective(x):
    residual = a @ x + b
    return jp.sum(0.5 * residual**2)

  pg = jaxopt.ProjectedGradient(
      objective,
      projection_non_negative,
      maxiter=sys.solver_iterations,
      implicit_diff=False,
      maxls=5,
  )

  qf_constraint = jac.T @ pg.run(jp.zeros_like(b)).params

  return qf_constraint



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers for Brax to support different upstream use cases."""

from typing import Dict, Optional

from brax.v2.envs import env as brax_env
from flax import struct
import jax
from jax import numpy as jp


def wrap_for_training(
    env: brax_env.Env, episode_length: int = 1000, action_repeat: int = 1
) -> brax_env.Wrapper:
  """Common wrapper pattern for all training agents.

  Args:
    env: environment to be wrapped
    episode_length: length of episode
    action_repeat: how many repeated actions to take per step

  Returns:
    An environment that is wrapped with Episode and AutoReset wrappers.  If the
    environment did not already have batch dimensions, it is additional Vmap
    wrapped.
  """
  env = EpisodeWrapper(env, episode_length, action_repeat)
  env = VmapWrapper(env)
  env = AutoResetWrapper(env)
  return env


class VmapWrapper(brax_env.Wrapper):
  """Vectorizes Brax env."""

  def __init__(self, env: brax_env.Env, batch_size: Optional[int] = None):
    super().__init__(env)
    self.batch_size = batch_size

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    if self.batch_size is not None:
      rng = jax.random.split(rng, self.batch_size)
    return jax.vmap(self.env.reset)(rng)

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    return jax.vmap(self.env.step)(state, action)


class EpisodeWrapper(brax_env.Wrapper):
  """Maintains episode step count and sets done at episode end."""

  def __init__(
      self, env: brax_env.Env, episode_length: int, action_repeat: int
  ):
    super().__init__(env)
    self.episode_length = episode_length
    self.action_repeat = action_repeat

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['steps'] = jp.zeros(rng.shape[:-1])
    state.info['truncation'] = jp.zeros(rng.shape[:-1])
    return state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    def f(state, _):
      nstate = self.env.step(state, action)
      return nstate, nstate.reward

    state, rewards = jax.lax.scan(f, state, (), self.action_repeat)
    state = state.replace(reward=jp.sum(rewards, axis=0))
    steps = state.info['steps'] + self.action_repeat
    one = jp.ones_like(state.done)
    zero = jp.zeros_like(state.done)
    episode_length = jp.array(self.episode_length, dtype=jp.int32)
    done = jp.where(steps >= episode_length, one, state.done)
    state.info['truncation'] = jp.where(
        steps >= episode_length, 1 - state.done, zero
    )
    state.info['steps'] = steps
    return state.replace(done=done)


class AutoResetWrapper(brax_env.Wrapper):
  """Automatically resets Brax envs that are done."""

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['first_pipeline_state'] = state.pipeline_state
    state.info['first_obs'] = state.obs
    return state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    if 'steps' in state.info:
      steps = state.info['steps']
      steps = jp.where(state.done, jp.zeros_like(steps), steps)
      state.info.update(steps=steps)
    state = state.replace(done=jp.zeros_like(state.done))
    state = self.env.step(state, action)

    def where_done(x, y):
      done = state.done
      if done.shape:
        done = jp.reshape(done, [x.shape[0]] + [1] * (len(x.shape) - 1))  # type: ignore
      return jp.where(done, x, y)

    pipeline_state = jax.tree_map(
        where_done, state.info['first_pipeline_state'], state.pipeline_state
    )
    obs = where_done(state.info['first_obs'], state.obs)
    return state.replace(pipeline_state=pipeline_state, obs=obs)


@struct.dataclass
class EvalMetrics:
  """Dataclass holding evaluation metrics for Brax.

  Attributes:
      episode_metrics: Aggregated episode metrics since the beginning of the
        episode.
      active_episodes: Boolean vector tracking which episodes are not done yet.
      episode_steps: Integer vector tracking the number of steps in the episode.
  """

  episode_metrics: Dict[str, jp.ndarray]
  active_episodes: jp.ndarray
  episode_steps: jp.ndarray


class EvalWrapper(brax_env.Wrapper):
  """Brax env with eval metrics."""

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    reset_state = self.env.reset(rng)
    reset_state.metrics['reward'] = reset_state.reward
    eval_metrics = EvalMetrics(
        episode_metrics=jax.tree_util.tree_map(
            jp.zeros_like, reset_state.metrics
        ),
        active_episodes=jp.ones_like(reset_state.reward),
        episode_steps=jp.zeros_like(reset_state.reward),
    )
    reset_state.info['eval_metrics'] = eval_metrics
    return reset_state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    state_metrics = state.info['eval_metrics']
    if not isinstance(state_metrics, EvalMetrics):
      raise ValueError(
          f'Incorrect type for state_metrics: {type(state_metrics)}'
      )
    del state.info['eval_metrics']
    nstate = self.env.step(state, action)
    nstate.metrics['reward'] = nstate.reward
    episode_steps = jp.where(
        state_metrics.active_episodes,
        nstate.info['steps'],
        state_metrics.episode_steps,
    )
    episode_metrics = jax.tree_util.tree_map(
        lambda a, b: a + b * state_metrics.active_episodes,
        state_metrics.episode_metrics,
        nstate.metrics,
    )
    active_episodes = state_metrics.active_episodes * (1 - nstate.done)

    eval_metrics = EvalMetrics(
        episode_metrics=episode_metrics,
        active_episodes=active_episodes,
        episode_steps=episode_steps,
    )
    nstate.info['eval_metrics'] = eval_metrics
    return nstate



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the core environment with some extra statistics for training.

*This is branched from training/env.py, and will be folded back later.*
"""

from typing import Callable, Dict, Tuple
from brax import envs
import flax
import jax
import jax.numpy as jnp


@flax.struct.dataclass
class EnvState:
  """Contains training state for the learner."""
  core: envs.State
  total_metrics: Dict[str, jnp.ndarray]
  total_episodes: jnp.ndarray


Action = jnp.ndarray
StepFn = Callable[[EnvState, Action], EnvState]


def wrap(core_env: envs.Env,
         rng: jnp.ndarray,
         extra_step_kwargs: bool = True) -> Tuple[EnvState, StepFn]:
  """Returns a wrapped state and step function for training."""
  first_core = jax.jit(core_env.reset)(rng)
  first_core.metrics['reward'] = first_core.reward
  first_core.metrics.update(
      {f'reward/{k}': v for k, v in first_core.info.get('rewards', {}).items()})
  first_core.metrics['score'] = first_core.info.get(
      'score', jnp.zeros_like(first_core.reward))
  first_core.metrics.update(
      {f'score/{k}': v for k, v in first_core.info.get('scores', {}).items()})
  first_total_metrics = jax.tree_util.tree_map(jnp.sum, first_core.metrics)
  first_total_episodes = jnp.zeros(())

  first_state = EnvState(
      core=first_core,
      total_metrics=first_total_metrics,
      total_episodes=first_total_episodes)

  def step(state: EnvState,
           action: Action,
           normalizer_params: Dict[str, jnp.ndarray] = None,
           extra_params: Dict[str, Dict[str, jnp.ndarray]] = None) -> EnvState:
    if extra_step_kwargs:
      core = core_env.step(
          state.core,
          action,
          normalizer_params=normalizer_params,
          extra_params=extra_params)
    else:
      core = core_env.step(state.core, action)
    core.metrics['reward'] = core.reward
    core.metrics.update(
        {f'reward/{k}': v for k, v in core.info.get('rewards', {}).items()})
    core.metrics['score'] = core.info.get('score', jnp.zeros_like(core.reward))
    core.metrics.update(
        {f'score/{k}': v for k, v in core.info.get('scores', {}).items()})
    total_metrics = jax.tree_util.tree_map(lambda a, b: a + jnp.sum(b),
                                      state.total_metrics, core.metrics)
    total_episodes = state.total_episodes + jnp.sum(core.done)
    return EnvState(
        core=core, total_metrics=total_metrics, total_episodes=total_episodes)

  return first_state, jax.jit(step)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""HalfCheetah that jumps."""

from brax import jumpy as jp
from brax.envs import env
from brax.envs import halfcheetah


class JumpCheetah(halfcheetah.Halfcheetah):
  """Run + Jump HalfCheetah."""

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    qpos = self.sys.default_angle() + jp.random_uniform(
        rng1, (self.sys.num_joint_dof,), -.1, .1)
    qvel = jp.random_uniform(rng2, (self.sys.num_joint_dof,), -.1, .1)
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_ctrl_cost': zero,
        'reward_forward': zero,
        'reward_hop': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    x_before = state.qp.pos[0, 0]
    x_after = qp.pos[0, 0]
    forward_reward = (x_after - x_before) / self.sys.config.dt
    ctrl_cost = -.1 * jp.sum(jp.square(action))
    z_before = state.qp.pos[0, 2]
    z_after = qp.pos[0, 2]
    hop_reward = jp.abs(z_after - z_before) / self.sys.config.dt
    reward = forward_reward + ctrl_cost + hop_reward
    state.metrics.update(
        reward_ctrl_cost=ctrl_cost,
        reward_forward=forward_reward,
        reward_hop=hop_reward)

    return state.replace(qp=qp, obs=obs, reward=reward)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Function to resolve collisions."""
# pylint:disable=g-multiple-import
from typing import Optional, Tuple

from brax.v2 import math
from brax.v2.base import Contact, Motion, System, Transform
import jax
from jax import numpy as jp
from jax import tree_map


def resolve(
    sys: System,
    xi: Transform,
    xdi: Motion,
    inv_inertia: jp.ndarray,
    contact: Optional[Contact],
) -> Tuple[Motion, jp.ndarray, jp.ndarray]:
  """Resolves springy collision constraint.

  Args:
    sys: System to forward propagate
    xi: Transform state of link center of mass
    xdi: Motion state of link center of mass
    inv_inertia: inverse inertia tensor at the center of mass in world frame
    contact: Contact pytree

  Returns:
    Tuple of
    p: world space impulse to apply to each link
    positions: location in world space to apply each impulse
    idxs: link to which impulse is applied
  """
  if contact is None:
    return Motion.zero((1,)), jp.array([]), jp.array([])

  @jax.vmap
  def impulse(contact):
    link_idx = jp.array(contact.link_idx)
    rel_pos = contact.pos - xi.take(link_idx).pos
    xd = xdi.take(link_idx)
    rel_vel = xd.vel + jax.vmap(jp.cross)(xd.ang, rel_pos)
    rel_vel *= (link_idx > -1).reshape(-1, 1)
    contact_vel = rel_vel[0] - rel_vel[1]
    normal_vel = jp.dot(contact.normal, contact_vel)

    link = sys.link.take(link_idx)
    i_inv = inv_inertia.take(link_idx, axis=0)
    i_inv *= (link_idx > -1).reshape(-1, 1, 1)

    temp1 = i_inv[0] @ jp.cross(rel_pos[0], contact.normal)
    temp2 = i_inv[1] @ jp.cross(rel_pos[1], contact.normal)
    ang = jp.dot(
        contact.normal,
        jp.cross(temp1, rel_pos[0]) + jp.cross(temp2, rel_pos[1]),
    )
    invmass = (1 / link.inertia.mass) * (link_idx > -1)
    denom = invmass[0] + invmass[1] + ang
    baumgarte_vel = sys.baumgarte_erp / sys.dt * contact.penetration
    impulse = (
        -1.0 * (1.0 + contact.elasticity) * normal_vel + baumgarte_vel
    ) / denom
    impulse_vec = impulse * contact.normal

    # apply drag due to friction acting parallel to the surface contact
    vel_d = contact_vel - normal_vel * contact.normal
    dir_d = vel_d / (1e-6 + math.safe_norm(vel_d))
    temp1 = i_inv[0] @ jp.cross(rel_pos[0], dir_d)
    temp2 = i_inv[1] @ jp.cross(rel_pos[1], dir_d)
    ang_d = jp.dot(
        dir_d, jp.cross(temp1, rel_pos[0]) + jp.cross(temp2, rel_pos[1])
    )
    impulse_d = math.safe_norm(vel_d) / (invmass[0] + invmass[1] + ang_d)

    # drag magnitude cannot exceed max friction
    impulse_d = jp.minimum(impulse_d, contact.friction * impulse)
    impulse_d_vec = -1.0 * impulse_d * dir_d

    # apply collision if penetrating, approaching, and oriented correctly
    apply_n = (contact.penetration >= 0.0) & (normal_vel < 0) & (impulse > 0.0)
    # apply drag if moving laterally above threshold
    apply_d = apply_n * (math.safe_norm(vel_d) > 1e-3)

    return impulse_vec * apply_n + impulse_d_vec * apply_d, contact.pos

  p, pos = impulse(contact)

  link_idx = jp.concatenate(contact.link_idx)
  p = tree_map(lambda *t: jp.concatenate(t), p, -p)
  p *= link_idx.reshape((-1, 1)) != -1
  p = Motion.create(vel=p)
  pos = jp.tile(pos, (2, 1))

  return p, pos, link_idx



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common types used throughout Acme.

This file was taken from acme and modified to simplify dependencies:

https://github.com/deepmind/acme/blob/master/acme/types.py
"""
from typing import Any, Iterable, Mapping, Union

from brax.training.acme import specs
import jax.numpy as jnp

# Define types for nested arrays and tensors.
NestedArray = jnp.ndarray
NestedTensor = Any

# pytype: disable=not-supported-yet
NestedSpec = Union[
    specs.Array,
    Iterable['NestedSpec'],
    Mapping[Any, 'NestedSpec'],
]
# pytype: enable=not-supported-yet

Nest = Union[NestedArray, NestedTensor, NestedSpec]



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Physics pipeline for generalized coordinates engine."""

from typing import Optional

from brax.v2 import actuator
from brax.v2 import base
from brax.v2 import geometry
from brax.v2 import kinematics
from brax.v2 import math
from brax.v2.base import Contact, Inertia, Motion, System
from brax.v2.generalized import constraint
from brax.v2.generalized import dynamics
from brax.v2.generalized import integrator
from brax.v2.generalized import mass
from flax import struct
import jax
from jax import numpy as jp


@struct.dataclass
class State(base.State):
  """Dynamic state that changes after every step.

  Attributes:
    com: center of mass position
    cinr: inertia in com frame
    cd: body velocities in com frame
    cdof: dofs in com frame
    cdofd: cdof velocity
    mass_mx: (qd_size, qd_size) mass matrix
    mass_mx_inv: (qd_size, qd_size) inverse mass matrix
    contact: calculated contacts
    con_jac: constraint jacobian
    con_pos: constraint position
    con_diag: constraint A diagonal
    qf_smooth: smooth dynamics force
    qf_constraint: (qd_size,) force from constraints (collision etc)
    qdd: (qd_size,) joint acceleration vector
  """

  # position/velocity based terms are updated at the end of each step:
  com: jp.ndarray
  cinr: Inertia
  cd: Motion
  cdof: Motion
  cdofd: Motion
  mass_mx: jp.ndarray
  mass_mx_inv: jp.ndarray
  contact: Optional[Contact]
  con_jac: jp.ndarray
  con_pos: jp.ndarray
  con_diag: jp.ndarray
  # acceleration based terms are calculated using terms from the previous step:
  qf_smooth: jp.ndarray
  qf_constraint: jp.ndarray
  qdd: jp.ndarray


def init(sys: System, q: jp.ndarray, qd: jp.ndarray) -> State:
  """Initializes physics state.

  Args:
    sys: a brax system
    q: (q_size,) joint angle vector
    qd: (qd_size,) joint velocity vector

  Returns:
    state: initial physics state
  """
  x, xd = kinematics.forward(sys, q, qd)
  com, cinr, cd, cdof, cdofd = dynamics.transform_com(sys, q, qd, x)
  mass_mx = mass.matrix(sys, cinr, cdof)
  one = jp.eye(sys.qd_size())
  mass_mx_inv = jax.scipy.linalg.solve(mass_mx, one, assume_a='pos')
  contact = geometry.contact(sys, x)
  con_jac, con_pos, con_diag = constraint.jacobian(sys, q, com, cdof, contact)
  qf_smooth, qf_constraint, qdd = jp.zeros((3, sys.qd_size()))

  return State(
      q,
      qd,
      x,
      xd,
      com,
      cinr,
      cd,
      cdof,
      cdofd,
      mass_mx,
      mass_mx_inv,
      contact,
      con_jac,
      con_pos,
      con_diag,
      qf_smooth,
      qf_constraint,
      qdd,
  )


def step(sys: System, state: State, act: jp.ndarray) -> State:
  """Performs a physics step.

  Args:
    sys: a brax system
    state: physics state prior to step
    act: (act_size,) actuator input vector

  Returns:
    state: physics state after step
  """
  tau = actuator.to_tau(sys, act, state.q)

  # calculate acceleration terms
  qf_smooth = dynamics.forward(
      sys, state.q, state.qd, state.cinr, state.cd, state.cdof, state.cdofd, tau
  )
  qf_constraint = constraint.force(
      sys,
      state.qd,
      qf_smooth,
      state.mass_mx_inv,
      state.con_jac,
      state.con_pos,
      state.con_diag,
  )
  # add dof damping to the mass matrix
  # because we already have M^-1, we use the derivative of the inverse:
  # (A +  εX)^-1 = A^-1 - εA^-1 @ X @ A^-1 + O(ε^2)
  mx_inv = state.mass_mx_inv
  mx_inv_damp = mx_inv - mx_inv @ (jp.diag(sys.dof.damping) * sys.dt) @ mx_inv
  qdd = mx_inv_damp @ (qf_smooth + qf_constraint)

  # update position/velocity level terms
  q, qd = integrator.integrate(sys, state.q, state.qd, qdd)
  x, xd = kinematics.forward(sys, q, qd)
  com, cinr, cd, cdof, cdofd = dynamics.transform_com(sys, q, qd, x)
  mass_mx = mass.matrix(sys, cinr, cdof)
  mass_mx_inv = math.inv_approximate(mass_mx, state.mass_mx_inv)
  contact = geometry.contact(sys, x)
  con_jac, con_pos, con_diag = constraint.jacobian(sys, q, com, cdof, contact)

  return State(
      q,
      qd,
      x,
      xd,
      com,
      cinr,
      cd,
      cdof,
      cdofd,
      mass_mx,
      mass_mx_inv,
      contact,
      con_jac,
      con_pos,
      con_diag,
      qf_smooth,
      qf_constraint,
      qdd,
  )



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for actuators."""

from absl.testing import absltest
from brax.v2 import actuator
from brax.v2 import test_utils
from brax.v2.generalized import pipeline
from jax import numpy as jp
import mujoco
import numpy as np

assert_almost_equal = np.testing.assert_almost_equal


# TODO: run actuator tests over all backend pipelines
def _actuator_step(sys, q, qd, dt, act):
  sys = sys.replace(dt=dt)
  state = pipeline.step(sys, pipeline.init(sys, q, qd), act)
  return state.q, state.qd


class ActuatorTest(absltest.TestCase):

  def test_motor(self):
    """Tests a single pendulum with motor actuator."""
    sys = test_utils.load_fixture('single_pendulum_motor.xml')
    mj_model = test_utils.load_fixture_mujoco('single_pendulum_motor.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)

    tau = jp.array([0.5 * 9.81])  # -mgl sin(theta)
    act = jp.array([1.0 / 150.0 * 0.5 * 9.81])
    tau2 = actuator.to_tau(sys, act, q)
    np.testing.assert_array_almost_equal(tau, tau2, 5)

    q2, qd2 = _actuator_step(sys, q, qd, dt=0.01, act=act)
    np.testing.assert_array_almost_equal(q2, jp.array([0]), 5)
    np.testing.assert_array_almost_equal(qd2, jp.array([0]), 5)

  def test_position(self):
    """Tests a single pendulum with position actuator."""
    sys = test_utils.load_fixture('single_pendulum_position.xml')
    mj_model = test_utils.load_fixture_mujoco('single_pendulum_position.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)
    theta = jp.pi / 2.0  # pendulum is vertical
    q = jp.array([theta])

    # a position actuator at the bottom should not move the pendulum
    act = jp.array([theta])
    tau = actuator.to_tau(sys, act, q)
    np.testing.assert_array_almost_equal(tau, jp.array([0]), 5)

    # put the pendulum into the horizontal position with the positional actuator
    # we know alpha = -2 theta/t^2, and I * alpha = (act-bias) * gear, and we
    # then solve for act. Semi-implicit Euler leaves off a factor of two.
    act = jp.array([-(theta * 0.5**2) / (0.01**2 * 10.0) + theta])
    q2, _ = _actuator_step(sys, q, qd, dt=0.01, act=act)
    np.testing.assert_array_almost_equal(q2, jp.array([0]), 1)

  def test_three_link_pendulum(self):
    """Tests a three link pendulum with a motor actuator."""
    sys = test_utils.load_fixture('triple_pendulum_motor.xml')
    mj_model = test_utils.load_fixture_mujoco('triple_pendulum_motor.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)
    theta = jp.pi / 2.0  # pendulum is vertical
    q = jp.array([theta, 0.0, 0.0])

    # test that with zero action the triple pendulum does not move
    act = jp.array([0] * 3)
    q1, qd1 = _actuator_step(sys, q, qd, dt=0.01, act=act)
    np.testing.assert_array_almost_equal(q1, q, 2)
    np.testing.assert_array_almost_equal(qd1, jp.zeros_like(qd1), 2)

    # test that each torque results in directionally accurate q and qd
    act = 1.0 / 150.0 * jp.array([0, -10, 10])
    q2, qd2 = _actuator_step(sys, q, qd, dt=0.001, act=act)
    self.assertAlmostEqual(q2[0], q[0], 3)
    self.assertLess(q2[1], q[1])
    self.assertGreater(q2[2], q[2])
    self.assertLess(qd2[1], -0.2)
    self.assertGreater(qd2[2], -0.2)

  def test_spherical_pendulum(self):
    """Tests a spherical pendulum with a position actuator."""
    sys = test_utils.load_fixture('single_spherical_pendulum_position.xml')
    mj_model = test_utils.load_fixture_mujoco(
        'single_spherical_pendulum_position.xml'
    )
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)

    act = jp.array([2.0, 3.0, 1.0])
    tau = actuator.to_tau(sys, act, q)
    expected_tau = jp.array([act[1], act[2], act[0]]) * 10.0
    np.testing.assert_array_almost_equal(tau, expected_tau)

    mj_data.ctrl = act
    mujoco.mj_step(mj_model, mj_data)
    expected_q, expected_qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)

    q2, qd2 = _actuator_step(sys, q, qd, dt=sys.dt, act=act)
    np.testing.assert_array_almost_equal(q2, expected_q)
    np.testing.assert_array_almost_equal(qd2, expected_qd)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environments.

In Braxlines Composer, all environments are defined through
a descriotion dictionary `env_desc`.

composer.py loads from `ENV_DESCS` with `env_name`,
where each entry can be a `env_desc` or a function that returns `env_desc`.
"""
# pylint:disable=protected-access
# pylint:disable=g-complex-comprehension
import copy
import functools
import importlib
import inspect
from typing import Any, Dict

ENV_DESCS = {}
DEFAULT_REGISTER_LIBS = ('sa_descs', 'ma_descs')


def is_env_desc(env_desc: Any):
  """Check if it is appropriate env_desc object."""
  return isinstance(env_desc, dict) or callable(env_desc)


def register_env(env_name: str, env_desc: Any, override: bool = True):
  """Register env_name and return env_descs."""
  global ENV_DESCS
  if not override and env_name in ENV_DESCS:
    return copy.deepcopy(ENV_DESCS[env_name])
  else:
    ENV_DESCS[env_name] = env_desc
    return copy.deepcopy(ENV_DESCS[env_name])


def register_lib(load_path: str):
  """Register a library of env_names with env_descs."""
  global ENV_DESCS
  if '.' not in load_path:
    load_path = f'brax.experimental.composer.envs.{load_path}'
  env_lib = importlib.import_module(load_path)
  ENV_DESCS.update(env_lib.ENV_DESCS)


def register_default_libs():
  """Register all default env."""
  for load_path in DEFAULT_REGISTER_LIBS:
    register_lib(load_path)


def list_env():
  """List registered environments."""
  return sorted(ENV_DESCS)


def exists(env_name: str):
  """If environment is registered."""
  return env_name in ENV_DESCS


def get_func_kwargs(func):
  """Get keyword args of a function."""
  # first, unwrap functools.partial. only extra keyword arguments.
  partial_supported_params = {}
  while isinstance(func, functools.partial):
    partial_supported_params.update(func.keywords)
    func = func.func
  # secondly, inspect the original function for keyword arguments.
  fn_params = inspect.signature(func).parameters
  support_kwargs = any(
      v.kind == inspect.Parameter.VAR_KEYWORD for v in fn_params.values())
  supported_params = {
      k: v.default
      for k, v in fn_params.items()
      if v.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD and
      v.default != inspect._empty
  }
  supported_params.update(partial_supported_params)
  return supported_params, support_kwargs


def inspect_env(env_name: str):
  """Inspect parameters of the env."""
  desc = env_name
  if isinstance(desc, str):
    desc = ENV_DESCS[desc]
  assert callable(desc) or isinstance(desc, dict), desc
  if not callable(desc):
    return {}, False
  return get_func_kwargs(desc)


def assert_env_params(env_name: str,
                      env_params: Dict[str, Any],
                      ignore_kwargs: bool = False):
  """Assert env_params are valid parameters for env_name."""
  assert isinstance(env_params, dict), env_params
  supported_params, support_kwargs = inspect_env(env_name)
  # if unnamed **kwargs, then always assert True
  if support_kwargs and not ignore_kwargs:
    return
  assert all(
      k in supported_params
      for k in env_params), f'invalid {env_params} for {supported_params}'



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.envs.grasp."""

from absl.testing import absltest
from brax import envs
from brax import jumpy as jp
import jax


class GraspTest(absltest.TestCase):

  def testGrasp(self):
    env = envs.create('grasp')
    grasp_action = jp.array([
     -.3, 0., 0., -1, 0., 0., -1, -0, 0.,  # gripper arm 1
      .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,     # gripper arm 2
      .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,      # gripper arm 3
      .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,      # gripper arm 4
      0, 0, -1.0           # position action
      ])

    def raise_action(i):
      return jp.array([
         -.3, 0., 0., -1, 0., 0., -1, -0, 0.,  # gripper arm 1
          .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,     # gripper arm 2
          .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,      # gripper arm 3
          .3, 0., 0.,  1, 0., 0.,  1,  0, 0.,      # gripper arm 4
        0, 0, -1.0*((250 - i)/250) + -.5*(i/250)           # position action
      ])

    state = env.reset(jp.random_prngkey(0))
    step = jax.jit(env.step)
    # grasp
    for _ in range(250):
      state = step(state, grasp_action)
    # slowly lift
    for i in range(250):
      state = step(state, raise_action(i))

    self.assertGreater(state.qp.pos[1, 2], 1.36)  # ball lifted off ground
    self.assertLess(state.qp.ang[1, 2], .01)  # ball not rolling


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Print the Center of Mass of a mesh file."""

from collections.abc import Sequence

from absl import app
from absl import flags
from brax.io import file
from trimesh.exchange.load import load_mesh

_PATH = flags.DEFINE_string('path', None, 'Path to mesh file.')


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  with file.File(_PATH.value, 'rb') as f:
    mesh = load_mesh(f, file_type=_PATH.value)
  if not mesh.is_watertight:
    raise AssertionError('Mesh must be watertight.')
  print(
      f'{{ x: {mesh.center_mass[0]} y: {mesh.center_mass[1]}  z: {mesh.center_mass[2]} }}'
  )


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a swimmer to swim in the +x direction."""

import brax
from brax import jumpy as jp
from brax import math
from brax.envs import env


class Swimmer(env.Env):



  """
  ### Description

  This environment corresponds to the Swimmer environment described in Rémi
  Coulom's PhD thesis
  ["Reinforcement Learning Using Neural Networks, with Applications to Motor Control"](https://tel.archives-ouvertes.fr/tel-00003985/document).

  The environment aims to increase the number of independent state and control
  variables as compared to the classic control environments. The swimmers
  consist of three or more segments ('***links***') and one less articulation
  joints ('***rotors***') - one rotor joint connecting exactly two links to
  form a linear chain.

  The swimmer is suspended in a two dimensional pool and always starts in the
  same position (subject to some deviation drawn from an uniform distribution),
  and the goal is to move as fast as possible towards the right by applying
  torque on the rotors and using the fluids friction.

  ### Notes

  The problem parameters are:

  * *n*: number of body parts
  * *m<sub>i</sub>*: mass of part *i* (*i* ∈ {1...n})
  * *l<sub>i</sub>*: length of part *i* (*i* ∈ {1...n})
  * *k*: viscous-friction coefficient

  While the default environment has *n* = 3, *l<sub>i</sub>* = 0.1,
  and *k* = 0.1. It is possible to tweak the MuJoCo XML files to increase the
  number of links, or to tweak any of the parameters.

  ### Action Space

  The agent take a 2-element vector for actions. The action space is a
  continuous `(action, action)` in `[-1, 1]`, where `action` represents the
  numerical torques applied between *links*

  | Num | Action                             | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied on the first rotor  | -1          | 1           | rot2                           | hinge | torque (N m) |
  | 1   | Torque applied on the second rotor | -1          | 1           | rot3                           | hinge | torque (N m) |

  ### Observation Space

  The state space consists of:

  * A<sub>0</sub>: position of first point
  * θ<sub>i</sub>: angle of part *i* with respect to the *x* axis
  * A<sub>0</sub>, θ<sub>i</sub>: their derivatives with respect to time (velocity and angular velocity)

  The observation is a `ndarray` with shape `(12,)` where the elements correspond to the following:

  | Num | Observation                          | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|---------------------------------------|------|-----|-------------------------------|-------|--------------------------|
  | 0   | angle of the front tip                | -Inf | Inf | rot                           | hinge | angle (rad)              |
  | 1   | angle of the second rotor             | -Inf | Inf | rot2                          | hinge | angle (rad)              |
  | 2   | angle of the second rotor             | -Inf | Inf | rot3                          | hinge | angle (rad)              |
  | 3   | velocity of the tip along the x-axis  | -Inf | Inf | slider1                       | slide | velocity (m/s)           |
  | 4   | velocity of the tip along the y-axis  | -Inf | Inf | slider2                       | slide | velocity (m/s)           |
  | 5   | velocity of the mid along the x-axis  | -Inf | Inf | N/A                           | slide | velocity (m/s)           |
  | 6   | velocity of the back along the y-axis | -Inf | Inf | N/A                           | slide | velocity (m/s)           |
  | 7   | velocity of the back along the x-axis | -Inf | Inf | N/A                           | slide | velocity (m/s)           |
  | 8   | velocity of the tip along the y-axis  | -Inf | Inf | N/A                           | slide | velocity (m/s)           |
  | 9   | angular velocity of front tip         | -Inf | Inf | rot                           | hinge | angular velocity (rad/s) |
  | 10  | angular velocity of second rotor      | -Inf | Inf | rot2                          | hinge | angular velocity (rad/s) |
  | 11  | angular velocity of third rotor       | -Inf | Inf | rot3                          | hinge | angular velocity (rad/s) |

  ### Rewards

  The reward consists of two parts:

  - *reward_fwd*: A reward of moving forward which is measured as
    *(x-coordinate before action - x-coordinate after action) / dt*. *dt* is the
    time between actions - the default *dt = 0.04*. This reward would be positive
    if the swimmer swims right as desired.
  - *reward_control*: A negative reward for penalising the swimmer if it takes
    actions that are too large. It is measured as *-coefficient x
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.0001

  ### Starting State

  All observations start in state (0,0,0,0,0,0,0,0) with a Uniform noise in the
  range of [-0.1, 0.1] is added to the initial state for stochasticity.

  ### Episode Termination

  The episode terminates when the episode length is greater than 1000.

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder
  (or by changing the path to a modified XML file in another folder).

  ```
  gym.make('Swimmer-v2')
  ```

  v3 and v4 take gym.make kwargs such as ctrl_cost_weight, reset_noise_scale
  etc.

  ```
  env = gym.make('Swimmer-v4', ctrl_cost_weight=0.1, ....)
  ```

  And a v5 version that uses Brax:

  ```
  env = gym.make('Swimmer-v5')
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as ctrl_cost_weight, reset_noise_scale
    etc. rgb rendering comes from tracking camera (so agent does not run away
    from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks. Added
    reward_threshold to environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self,
               forward_reward_weight=1.0,
               ctrl_cost_weight=1e-4,
               reset_noise_scale=0.1,
               exclude_current_positions_from_observation=True,
               legacy_reward=False,
               legacy_spring=False,
               **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

    self._forward_reward_weight = forward_reward_weight
    self._ctrl_cost_weight = ctrl_cost_weight
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation)

    # these parameters were derived from the mujoco swimmer:
    viscosity = 0.1
    density = 10.0
    inertia = (
        0.17278759594743870,
        3.5709436495803999,
        3.5709436495803999,
    )
    body_mass = 34.557519189487735

    # convert inertia to box
    inertia = jp.array([
        inertia[1] + inertia[2] - inertia[0],
        inertia[0] + inertia[1] - inertia[2],
        inertia[0] + inertia[2] - inertia[1],
    ])
    inertia = jp.sqrt(inertia / (body_mass * 6))

    # spherical drag
    self._spherical_drag = -3 * jp.pi * jp.mean(inertia) * viscosity

    # corrections to spherical drag force due to shape of capsules
    self._fix_drag = 0.5 * density * jp.array([
        inertia[1] * inertia[2],
        inertia[0] * inertia[2],
        inertia[0] * inertia[1]
    ])

  def reset(self, rng: jp.ndarray) -> env.State:
    rng, rng1, rng2 = jp.random_split(rng, 3)
    qpos = self.sys.default_angle() + self._noise(rng1)
    qvel = self._noise(rng2)
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        "reward_fwd": zero,
        "reward_ctrl": zero,
        "x_position": zero,
        "y_position": zero,
        "distance_from_origin": zero,
        "x_velocity": zero,
        "y_velocity": zero,
        "forward_reward": zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    force = self._get_viscous_force(state.qp)
    act = jp.concatenate([action, force.reshape(-1)], axis=0)
    qp, info = self.sys.step(state.qp, act)

    com_before = self._center_of_mass(state.qp)
    com_after = self._center_of_mass(qp)
    velocity = (com_after - com_before) / self.sys.config.dt
    forward_reward = self._forward_reward_weight * velocity[0]
    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))

    obs = self._get_obs(qp, info)
    reward = forward_reward - ctrl_cost
    state.metrics.update(
        reward_fwd=forward_reward,
        reward_ctrl=-ctrl_cost,
        x_position=com_after[0],
        y_position=com_after[1],
        distance_from_origin=jp.norm(qp.pos[0]),
        x_velocity=velocity[0],
        y_velocity=velocity[1],
        forward_reward=forward_reward,
    )

    return state.replace(qp=qp, obs=obs, reward=reward)

  @property
  def action_size(self):
    return 2

  def _get_viscous_force(self, qp):
    """Calculate viscous force to apply to each body."""
    # ignore the floor
    qp = jp.take(qp, jp.arange(0, qp.vel.shape[0] - 1))

    # spherical drag force
    force = qp.vel * self._spherical_drag

    # corrections to spherical drag force due to shape of capsules
    vel = jp.vmap(math.rotate)(qp.vel, math.quat_inv(qp.rot))
    force -= jp.diag(self._fix_drag * jp.abs(vel) * vel)
    force = jp.vmap(math.rotate)(force, qp.rot)
    force = jp.clip(force, -5., 5.)

    return force

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Observe swimmer body position and velocities."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)

    # convert quaternion to rotation angle about z axis
    ang_z = math.quat_to_euler(qp.rot[0])[2:3]

    # qpos: position and orientation of the torso and the joint angles
    if self._exclude_current_positions_from_observation:
      qpos = [ang_z, joint_angle]
    else:
      qpos = [qp.pos[0, :2], ang_z, joint_angle]

    # # qvel: velocity of the bodies and the joint angle velocities
    qvel = [qp.vel[0, :2].ravel(), qp.ang[0, 2:], joint_vel]

    return jp.concatenate(qpos + qvel)

  def _center_of_mass(self, qp):
    """Returns the center of mass position of a swimmer with state qp."""
    mass, pos = self.sys.body.mass[:-1], qp.pos[:-1]
    return jp.sum(jp.vmap(jp.multiply)(mass, pos), axis=0) / jp.sum(mass)

  def _noise(self, rng):
    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    return jp.random_uniform(rng, (self.sys.num_joint_dof,), low, hi)


_SYSTEM_CONFIG = """
  bodies {
    name: "torso"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  bodies {
    name: "mid"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  bodies {
    name: "back"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 1
    frozen { all: true }
  }
  joints {
    name: "rot2"
    parent: "torso"
    child: "mid"
    parent_offset {
      x: 0.5
    }
    child_offset {
      x: -.5
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -100.0
      max: 100.0
    }
    angular_damping: 10.
    reference_rotation {
    }
  }
  joints {
    name: "rot3"
    parent: "mid"
    child: "back"
    parent_offset {
      x: .5
    }
    child_offset {
      x: -.5
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -100.0
      max: 100.0
    }
    angular_damping: 10.
    reference_rotation {
    }
  }
  actuators {
    name: "rot2"
    joint: "rot2"
    strength: 30.0
    torque {
    }
  }
  actuators {
    name: "rot3"
    joint: "rot3"
    strength: 30.0
    torque {
    }
  }
  forces {
    name: "torso_viscosity_thruster"
    body: "torso"
    strength: 1.0
    thruster {}
  }
  forces {
    name: "mid_viscosity_thruster"
    body: "mid"
    strength: 1.0
    thruster {}
  }
  forces {
    name: "back_viscosity_thruster"
    body: "back"
    strength: 1.0
    thruster {}
  }
  frozen {
    rotation { x: 1.0 y: 1.0 }
  }
  friction: 0.6
  angular_damping: -0.05
  collide_include { }
  dt: 0.04
  substeps: 12
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 1
    frozen { all: true }
  }
  bodies {
    name: "torso"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  bodies {
    name: "mid"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  bodies {
    name: "back"
    colliders {
      rotation {
        y: -90.0
      }
      capsule {
        radius: 0.1
        length: 1.2
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.35604717
  }
  joints {
    name: "rot2"
    stiffness: 10000.0
    parent: "torso"
    child: "mid"
    parent_offset {
      x: 0.5
    }
    child_offset {
      x: -.5
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -100.0
      max: 100.0
    }
    angular_damping: 10.
    reference_rotation {
    }
  }
  joints {
    name: "rot3"
    stiffness: 10000.0
    parent: "mid"
    child: "back"
    parent_offset {
      x: .5
    }
    child_offset {
      x: -.5
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -100.0
      max: 100.0
    }
    angular_damping: 10.
    reference_rotation {
    }
  }
  actuators {
    name: "rot2"
    joint: "rot2"
    strength: 30.0
    torque {
    }
  }
  actuators {
    name: "rot3"
    joint: "rot3"
    strength: 30.0
    torque {
    }
  }
  forces {
    name: "torso_viscosity_thruster"
    body: "torso"
    strength: 1.0
    thruster {}
  }
  forces {
    name: "mid_viscosity_thruster"
    body: "mid"
    strength: 1.0
    thruster {}
  }
  forces {
    name: "back_viscosity_thruster"
    body: "back"
    strength: 1.0
    thruster {}
  }
  frozen {
    rotation { x: 1.0 y: 1.0 }
  }
  friction: 0.6
  angular_damping: -0.05
  baumgarte_erp: 0.1
  collide_include { }
  dt: 0.04
  substeps: 12
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""composer Sweep."""
# pylint:disable=g-complex-comprehension
import itertools
from brax.experimental.braxlines.experiments import defaults

seed = 0
comps = [('ant', {}), ('pro_ant', dict(num_legs=[2, 6]))]

AGENT_MODULE = 'brax.experimental.composer.train'
CONFIG = [
    dict(
        env_name='chase',
        env_params=dict(
            main_agent=[comp1[0]],
            main_agent_params=comp1[1],
            other_agent=[comp2[0]],
            other_agent_params=comp2[1],
            num_agents=list(range(2, 5)),
        ),
        seed=seed,
        ppo_params=defaults.get_ppo_params(comp1[0], 3, default='ant'))
    for comp1, comp2 in itertools.product(comps, comps)
]



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax envs."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import envs
from brax.v2 import test_utils
from jax import numpy as jp

_EXPECTED_SPS = {
    'spring': {
        'ant': 1000,
    }
}



class EnvTest(parameterized.TestCase):
  params = [
      (b, e, _EXPECTED_SPS[b][e])
      for b in _EXPECTED_SPS
      for e in _EXPECTED_SPS[b]
  ]

  @parameterized.parameters(params)
  def testSpeed(self, backend, env_name, expected_sps):
    batch_size = 128
    episode_length = 1000

    env = envs.create(
        env_name,
        backend=backend,
        episode_length=episode_length,
        auto_reset=True,
    )
    zero_action = jp.zeros(env.action_size)
    step_fn = functools.partial(env.step, action=zero_action)

    mean_sps = test_utils.benchmark(
        f'{backend}_{env_name}',
        env.reset,
        step_fn,
        batch_size=batch_size,
        length=episode_length,
    )
    self.assertGreater(mean_sps, expected_sps * 0.99)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Soft Actor-Critic training.

See: https://arxiv.org/pdf/1812.05905.pdf
"""

import functools
import time
from typing import Any, Callable, Optional, Tuple

from absl import logging
from brax import envs
from brax.envs import wrappers
from brax.io import model
from brax.training import acting
from brax.training import gradients
from brax.training import pmap
from brax.training import replay_buffers
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.sac import losses as sac_losses
from brax.training.agents.sac import networks as sac_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax

Metrics = types.Metrics
Transition = types.Transition
InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]

ReplayBufferState = Any

_PMAP_AXIS_NAME = 'i'


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  policy_optimizer_state: optax.OptState
  policy_params: Params
  q_optimizer_state: optax.OptState
  q_params: Params
  target_q_params: Params
  gradient_steps: jnp.ndarray
  env_steps: jnp.ndarray
  alpha_optimizer_state: optax.OptState
  alpha_params: Params
  normalizer_params: running_statistics.RunningStatisticsState


def _unpmap(v):
  return jax.tree_util.tree_map(lambda x: x[0], v)


def _init_training_state(
    key: PRNGKey, obs_size: int, local_devices_to_use: int,
    sac_network: sac_networks.SACNetworks,
    alpha_optimizer: optax.GradientTransformation,
    policy_optimizer: optax.GradientTransformation,
    q_optimizer: optax.GradientTransformation) -> TrainingState:
  """Inits the training state and replicates it over devices."""
  key_policy, key_q = jax.random.split(key)
  log_alpha = jnp.asarray(0., dtype=jnp.float32)
  alpha_optimizer_state = alpha_optimizer.init(log_alpha)

  policy_params = sac_network.policy_network.init(key_policy)
  policy_optimizer_state = policy_optimizer.init(policy_params)
  q_params = sac_network.q_network.init(key_q)
  q_optimizer_state = q_optimizer.init(q_params)

  normalizer_params = running_statistics.init_state(
      specs.Array((obs_size,), jnp.float32))

  training_state = TrainingState(
      policy_optimizer_state=policy_optimizer_state,
      policy_params=policy_params,
      q_optimizer_state=q_optimizer_state,
      q_params=q_params,
      target_q_params=q_params,
      gradient_steps=jnp.zeros(()),
      env_steps=jnp.zeros(()),
      alpha_optimizer_state=alpha_optimizer_state,
      alpha_params=log_alpha,
      normalizer_params=normalizer_params)
  return jax.device_put_replicated(training_state,
                                   jax.local_devices()[:local_devices_to_use])


def train(environment: envs.Env,
          num_timesteps,
          episode_length: int,
          action_repeat: int = 1,
          num_envs: int = 1,
          num_eval_envs: int = 128,
          learning_rate: float = 1e-4,
          discounting: float = 0.9,
          seed: int = 0,
          batch_size: int = 256,
          num_evals: int = 1,
          normalize_observations: bool = False,
          max_devices_per_host: Optional[int] = None,
          reward_scaling: float = 1.,
          tau: float = 0.005,
          min_replay_size: int = 0,
          max_replay_size: Optional[int] = None,
          grad_updates_per_step: int = 1,
          deterministic_eval: bool = False,
          network_factory: types.NetworkFactory[
              sac_networks.SACNetworks] = sac_networks.make_sac_networks,
          progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
          checkpoint_logdir: Optional[str] = None,
          eval_env: Optional[envs.Env] = None):
  """SAC training."""
  process_id = jax.process_index()
  local_devices_to_use = jax.local_device_count()
  if max_devices_per_host is not None:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  device_count = local_devices_to_use * jax.process_count()
  logging.info('local_device_count: %s; total_device_count: %s',
               local_devices_to_use, device_count)

  if min_replay_size >= num_timesteps:
    raise ValueError(
        'No training will happen because min_replay_size >= num_timesteps')

  if max_replay_size is None:
    max_replay_size = num_timesteps

  # The number of environment steps executed for every `actor_step()` call.
  env_steps_per_actor_step = action_repeat * num_envs
  # equals to ceil(min_replay_size / env_steps_per_actor_step)
  num_prefill_actor_steps = -(-min_replay_size // num_envs)
  num_prefill_env_steps = num_prefill_actor_steps * env_steps_per_actor_step
  assert num_timesteps - num_prefill_env_steps >= 0
  num_evals_after_init = max(num_evals - 1, 1)
  # The number of run_one_sac_epoch calls per run_sac_training.
  # equals to
  # ceil(num_timesteps - num_prefill_env_steps /
  #      (num_evals_after_init * env_steps_per_actor_step))
  num_training_steps_per_epoch = -(
      -(num_timesteps - num_prefill_env_steps) //
      (num_evals_after_init * env_steps_per_actor_step))

  assert num_envs % device_count == 0
  env = environment
  env = wrappers.wrap_for_training(
      env, episode_length=episode_length, action_repeat=action_repeat)

  obs_size = env.observation_size
  action_size = env.action_size

  normalize_fn = lambda x, y: x
  if normalize_observations:
    normalize_fn = running_statistics.normalize
  sac_network = network_factory(
      observation_size=obs_size,
      action_size=action_size,
      preprocess_observations_fn=normalize_fn)
  make_policy = sac_networks.make_inference_fn(sac_network)

  alpha_optimizer = optax.adam(learning_rate=3e-4)

  policy_optimizer = optax.adam(learning_rate=learning_rate)
  q_optimizer = optax.adam(learning_rate=learning_rate)

  dummy_obs = jnp.zeros((obs_size,))
  dummy_action = jnp.zeros((action_size,))
  dummy_transition = Transition(
      observation=dummy_obs,
      action=dummy_action,
      reward=0.,
      discount=0.,
      next_observation=dummy_obs,
      extras={
          'state_extras': {
              'truncation': 0.
          },
          'policy_extras': {}
      })
  replay_buffer = replay_buffers.UniformSamplingQueue(
      max_replay_size=max_replay_size // device_count,
      dummy_data_sample=dummy_transition,
      sample_batch_size=batch_size * grad_updates_per_step // device_count)

  alpha_loss, critic_loss, actor_loss = sac_losses.make_losses(
      sac_network=sac_network,
      reward_scaling=reward_scaling,
      discounting=discounting,
      action_size=action_size)
  alpha_update = gradients.gradient_update_fn(
      alpha_loss, alpha_optimizer, pmap_axis_name=_PMAP_AXIS_NAME)
  critic_update = gradients.gradient_update_fn(
      critic_loss, q_optimizer, pmap_axis_name=_PMAP_AXIS_NAME)
  actor_update = gradients.gradient_update_fn(
      actor_loss, policy_optimizer, pmap_axis_name=_PMAP_AXIS_NAME)

  def sgd_step(
      carry: Tuple[TrainingState, PRNGKey],
      transitions: Transition) -> Tuple[Tuple[TrainingState, PRNGKey], Metrics]:
    training_state, key = carry

    key, key_alpha, key_critic, key_actor = jax.random.split(key, 4)

    alpha_loss, alpha_params, alpha_optimizer_state = alpha_update(
        training_state.alpha_params,
        training_state.policy_params,
        training_state.normalizer_params,
        transitions,
        key_alpha,
        optimizer_state=training_state.alpha_optimizer_state)
    alpha = jnp.exp(training_state.alpha_params)
    critic_loss, q_params, q_optimizer_state = critic_update(
        training_state.q_params,
        training_state.policy_params,
        training_state.normalizer_params,
        training_state.target_q_params,
        alpha,
        transitions,
        key_critic,
        optimizer_state=training_state.q_optimizer_state)
    actor_loss, policy_params, policy_optimizer_state = actor_update(
        training_state.policy_params,
        training_state.normalizer_params,
        training_state.q_params,
        alpha,
        transitions,
        key_actor,
        optimizer_state=training_state.policy_optimizer_state)

    new_target_q_params = jax.tree_util.tree_map(
        lambda x, y: x * (1 - tau) + y * tau, training_state.target_q_params,
        q_params)

    metrics = {
        'critic_loss': critic_loss,
        'actor_loss': actor_loss,
        'alpha_loss': alpha_loss,
        'alpha': jnp.exp(alpha_params),
    }

    new_training_state = TrainingState(
        policy_optimizer_state=policy_optimizer_state,
        policy_params=policy_params,
        q_optimizer_state=q_optimizer_state,
        q_params=q_params,
        target_q_params=new_target_q_params,
        gradient_steps=training_state.gradient_steps + 1,
        env_steps=training_state.env_steps,
        alpha_optimizer_state=alpha_optimizer_state,
        alpha_params=alpha_params,
        normalizer_params=training_state.normalizer_params)
    return (new_training_state, key), metrics

  def get_experience(
      normalizer_params: running_statistics.RunningStatisticsState,
      policy_params: Params, env_state: envs.State,
      buffer_state: ReplayBufferState, key: PRNGKey
  ) -> Tuple[running_statistics.RunningStatisticsState, envs.State,
             ReplayBufferState]:
    policy = make_policy((normalizer_params, policy_params))
    env_state, transitions = acting.actor_step(
        env, env_state, policy, key, extra_fields=('truncation',))

    normalizer_params = running_statistics.update(
        normalizer_params,
        transitions.observation,
        pmap_axis_name=_PMAP_AXIS_NAME)

    buffer_state = replay_buffer.insert(buffer_state, transitions)
    return normalizer_params, env_state, buffer_state

  def training_step(
      training_state: TrainingState, env_state: envs.State,
      buffer_state: ReplayBufferState, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, ReplayBufferState, Metrics]:
    experience_key, training_key = jax.random.split(key)
    normalizer_params, env_state, buffer_state = get_experience(
        training_state.normalizer_params, training_state.policy_params,
        env_state, buffer_state, experience_key)
    training_state = training_state.replace(
        normalizer_params=normalizer_params,
        env_steps=training_state.env_steps + env_steps_per_actor_step)

    buffer_state, transitions = replay_buffer.sample(buffer_state)
    # Change the front dimension of transitions so 'update_step' is called
    # grad_updates_per_step times by the scan.
    transitions = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, (grad_updates_per_step, -1) + x.shape[1:]),
        transitions)
    (training_state, _), metrics = jax.lax.scan(sgd_step,
                                                (training_state, training_key),
                                                transitions)

    metrics['buffer_current_size'] = buffer_state.current_size
    metrics['buffer_current_position'] = buffer_state.current_position
    return training_state, env_state, buffer_state, metrics

  def prefill_replay_buffer(
      training_state: TrainingState, env_state: envs.State,
      buffer_state: ReplayBufferState, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, ReplayBufferState, PRNGKey]:

    def f(carry, unused):
      del unused
      training_state, env_state, buffer_state, key = carry
      key, new_key = jax.random.split(key)
      new_normalizer_params, env_state, buffer_state = get_experience(
          training_state.normalizer_params, training_state.policy_params,
          env_state, buffer_state, key)
      new_training_state = training_state.replace(
          normalizer_params=new_normalizer_params,
          env_steps=training_state.env_steps + env_steps_per_actor_step)
      return (new_training_state, env_state, buffer_state, new_key), ()

    return jax.lax.scan(
        f, (training_state, env_state, buffer_state, key), (),
        length=num_prefill_actor_steps)[0]

  prefill_replay_buffer = jax.pmap(
      prefill_replay_buffer, axis_name=_PMAP_AXIS_NAME)

  def training_epoch(
      training_state: TrainingState, env_state: envs.State,
      buffer_state: ReplayBufferState, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, ReplayBufferState, Metrics]:

    def f(carry, unused_t):
      ts, es, bs, k = carry
      k, new_key = jax.random.split(k)
      ts, es, bs, metrics = training_step(ts, es, bs, k)
      return (ts, es, bs, new_key), metrics

    (training_state, env_state, buffer_state, key), metrics = jax.lax.scan(
        f, (training_state, env_state, buffer_state, key), (),
        length=num_training_steps_per_epoch)
    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    return training_state, env_state, buffer_state, metrics

  training_epoch = jax.pmap(training_epoch, axis_name=_PMAP_AXIS_NAME)

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(
      training_state: TrainingState, env_state: envs.State,
      buffer_state: ReplayBufferState, key: PRNGKey
  ) -> Tuple[TrainingState, envs.State, ReplayBufferState, Metrics]:
    nonlocal training_walltime
    t = time.time()
    (training_state, env_state, buffer_state,
     metrics) = training_epoch(training_state, env_state, buffer_state, key)
    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (env_steps_per_actor_step *
           num_training_steps_per_epoch) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()}
    }
    return training_state, env_state, buffer_state, metrics

  global_key, local_key = jax.random.split(jax.random.PRNGKey(seed))
  local_key = jax.random.fold_in(local_key, process_id)

  # Training state init
  training_state = _init_training_state(
      key=global_key,
      obs_size=obs_size,
      local_devices_to_use=local_devices_to_use,
      sac_network=sac_network,
      alpha_optimizer=alpha_optimizer,
      policy_optimizer=policy_optimizer,
      q_optimizer=q_optimizer)
  del global_key

  local_key, rb_key, env_key, eval_key = jax.random.split(local_key, 4)

  # Env init
  env_keys = jax.random.split(env_key, num_envs // jax.process_count())
  env_keys = jnp.reshape(env_keys,
                         (local_devices_to_use, -1) + env_keys.shape[1:])
  env_state = jax.pmap(env.reset)(env_keys)

  # Replay buffer init
  buffer_state = jax.pmap(replay_buffer.init)(
      jax.random.split(rb_key, local_devices_to_use))

  if not eval_env:
    eval_env = env
  else:
    eval_env = wrappers.wrap_for_training(
        eval_env, episode_length=episode_length, action_repeat=action_repeat)

  evaluator = acting.Evaluator(
      eval_env,
      functools.partial(make_policy, deterministic=deterministic_eval),
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key)

  # Run initial eval
  if process_id == 0 and num_evals > 1:
    metrics = evaluator.run_evaluation(
        _unpmap(
            (training_state.normalizer_params, training_state.policy_params)),
        training_metrics={})
    logging.info(metrics)
    progress_fn(0, metrics)

  # Create and initialize the replay buffer.
  t = time.time()
  prefill_key, local_key = jax.random.split(local_key)
  prefill_keys = jax.random.split(prefill_key, local_devices_to_use)
  training_state, env_state, buffer_state, _ = prefill_replay_buffer(
      training_state, env_state, buffer_state, prefill_keys)

  replay_size = jnp.sum(jax.vmap(
      replay_buffer.size)(buffer_state)) * jax.process_count()
  logging.info('replay size after prefill %s', replay_size)
  assert replay_size >= min_replay_size
  training_walltime = time.time() - t

  current_step = 0
  for _ in range(num_evals_after_init):
    logging.info('step %s', current_step)

    # Optimization
    epoch_key, local_key = jax.random.split(local_key)
    epoch_keys = jax.random.split(epoch_key, local_devices_to_use)
    (training_state, env_state, buffer_state,
     training_metrics) = training_epoch_with_timing(training_state, env_state,
                                                    buffer_state, epoch_keys)
    current_step = int(_unpmap(training_state.env_steps))

    # Eval and logging
    if process_id == 0:
      if checkpoint_logdir:
        # Save current policy.
        params = _unpmap(
            (training_state.normalizer_params, training_state.policy_params))
        path = f'{checkpoint_logdir}_sac_{current_step}.pkl'
        model.save_params(path, params)

      # Run evals.
      metrics = evaluator.run_evaluation(
          _unpmap(
              (training_state.normalizer_params, training_state.policy_params)),
          training_metrics)
      logging.info(metrics)
      progress_fn(current_step, metrics)

  total_steps = current_step
  assert total_steps >= num_timesteps

  params = _unpmap(
      (training_state.normalizer_params, training_state.policy_params))

  # If there was no mistakes the training_state should still be identical on all
  # devices.
  pmap.assert_is_replicated(training_state)
  logging.info('total steps: %s', total_steps)
  pmap.synchronize_hosts()
  return (make_policy, params, metrics)
